#!/usr/bin/perl -w

# Drall
# Copyright (C) 1997-2000 Henrik Edlund <henrik@edlund.org>
# http://www.edlund.org/hacks/drall/
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA.
#
# If you like Drall, please express your satisfaction with a donation: send me
# what you feel Drall has been worth to you. If you are glad that I developed
# Drall and distribute it as free software, rather than following the
# obstructive and antisocial practices typical of software developers, reward
# me. If you would like me to develop more free software, contribute.



use 5.004;
use strict;
use Fcntl qw(:DEFAULT :flock);

use vars qw($product_name $product_version);
$product_name = 'Drall';
$product_version = '1.3.2.3';



#################  Do NOT edit anything outside the Configuration if you don't
# Configuration #  know what you are doing. If you are planning to tweak then
#################  please read the License first.


use vars qw(%config);
%config =
    (


########################  This is the static configuration.
# Static Configuration #  This configuration is not to be changed by external
########################  or user configurations.


## This sets the support for multi configuration.
## Setting "0" means no support, the script will only use the normal
## configuration below and ignore any attempt to make it use an external
## config.
## Setting "1" means that the script will first read and initialize the
## configuration with the values in the normal configuration below and then
## reset the values with those found in the external configuration if any was
## specified as parameter when calling the script. If not the script will
## continue with the normal configuration.
## Setting any other value (for example "2") will make the script behave like
## with "1" but it will not fall back on normal configuration if an external
## was not specified. If so, an error will then be presented to the user.
## 0 - ignore multi configuration attempts
## 1 - support multi configuration but fall back to the normal if unspecified
## 2 - require multi configuration parameter

     'config_multi' => 0,


## Where to look for external configuration, when multi configuration is in
## use, and user configuration, when it is in use.
## External configurations should be named "1234.external.drall.pl" where
## "1234" is the name of the config specified as parameter when calling the
## script. The parameter is called "config" and is used for example by adding
## ";config=1234" to the URL for the script.
## User configurations should be named "username.user.drall.pl" where
## "username" is the authenticated username.
## Default below is "./" which is the same directory as the script is run
## from. An example external configuration named "1234.external.drall.pl" and
## an example user configuration named "username.user.drall.pl" can be found
## in the same archive as this script.

     'config_directory' => './',


#########################  This is the normal (dynamic) configuration. This
# Dynamic Configuration #  initializes all configuration variables. They can
#########################  be changed by an external or user configuration.


## Enable or disable the support for user configuration.
## If enabled and there is an authenticated user and an user configuration for
## this user then the config values will be reset with those found in the user
## configuration. If there is no user configuration the script will simply
## continue. See above for more about how to name user configurations.
## Note that as this is done after certain other operations you are not able
## to influence the values for the config variables "ssl", "authentication"
## and "suid" from an user configuration.

     'config_user' => 0,


## Force SSL connection. Script will not perform any action if it isn't run
## trough a secure (SSL) connection. Forcing this ("1") is recommended.

     'ssl' => 0,


## Enable (use) or disable the Averist authentication module (external).
## If enabled this requires a properly configured Averist module.
## If disabled (not in use) you can use HTTP authentication (configured in the
## web server), or no authentication at all (dangerous).
## If a correct authentication has been done using either of these two
## methods, then the !username! variable in the rest of the configuration will
## be replaced by the authenticated username.
## Averist authentication ("1") is recommended.

     'authentication' => 0,


## Set user id and group id. If there is an authenticated username available,
## and this username exists on the local (UNIX) system, the effective user id
## will be set to the local user id of this username, and the effective group
## id will be set to the local primary group id of this username.
## This requires that the script is owned by user "root" and has the "set user
## id" attribute set (chown root drall.pl; chmod u+s drall.pl). This is a
## potential SECURITY RISK so be sure you know what you are doing. You also
## need to change the first line in the script from "#!/usr/bin/perl -w" to
## either "#!/usr/bin/perl -U" or "#!/usr/bin/suidperl -U" depending on if
## your system allows "set user id" scripts directly or not.
## If this feature is enabled ("1") and the username does not exist on the
## local (UNIX) system the script will cause an error message to be displayed
## and proceed no further, to ensure that the script is not run as root.
## Note that if this feature is disabled ("0") and the script in changed as
## described above it will run as "root" all the time and that is a SECURITY
## RISK. Beware!

     'suid' => 0,


## All language phrases used in the script. No need to look at the code if you
## just want to translate this script to another language. Just translate the
## phrases below. !x! (where x is a number) are variables replaced at
## run-time.

     'language' =>
     [
      'Missing \'directory\' argument',                                  # 0
      'Missing \'file\' argument',                                       # 1
      'Illegal action specified',                                        # 2
      'Index of !0!',                                                    # 3
      'Parent directory',                                                # 4
      'Edit',                                                            # 5
      'Copy',                                                            # 6
      'Rename',                                                          # 7
      'Move',                                                            # 8
      'Delete',                                                          # 9
      'Upload file',                                                     # 10
      'Make directory',                                                  # 11
      'Delete report',                                                   # 12
      'Delete of directory !0! successful.',                             # 13
      'Delete of directory !0! failed.',                                 # 14
      'Delete of file !0! successful.',                                  # 15
      'Delete of file !0! failed.',                                      # 16
      'No files/directories were selected.',                             # 17
      'Return to file manager',                                          # 18
      'Couldn\'t retrieve file.',                                        # 19
      'Make of directory !0! successful.',                               # 20
      'Make of directory !0! failed.',                                   # 21
      'Error making directory.',                                         # 22
      'A file with the same name exists.',                               # 23
      'Directory already exists.',                                       # 24
      'Make a new directory in !0!',                                     # 25
      'Make directory',                                                  # 26
      'Upload of !0! to !1! successful.',                                # 27
      'Upload of !0! to !1! failed.',                                    # 28
      'Error writing to file.',                                          # 29
      'A directory of the same name exists.',                            # 30
      'File already exists - delete it or select to overwrite it.',      # 31
      'No file was selected.',                                           # 32
      'Upload file to !0!',                                              # 33
      'Make file name lowercase if uppercase is found in it',            # 34
      'Upload file',                                                     # 35
      'Make report',                                                     # 36
      'Upload report',                                                   # 37
      'Overwrite file if it already exists',                             # 38
      'Existing file overwritten.',                                      # 39
      'Bad config specified',                                            # 40
      'No config specified',                                             # 41
      'This function has been disabled',                                 # 42
      'Move of directory !0! to !1! successful.',                        # 43
      'Move of directory !0! to !1! failed.',                            # 44
      'Move of file !0! to !1! successful.',                             # 45
      'Move of file !0! to !1! failed.',                                 # 46
      'Move report',                                                     # 47
      'Rename files and directories in !0!',                             # 48
      'Rename',                                                          # 49
      'Rename of directory !0! to !1! successful.',                      # 50
      'Rename of directory !0! to !1! failed.',                          # 51
      'Rename of file !0! to !1! successful.',                           # 52
      'Rename of file !0! to !1! failed.',                               # 53
      'Rename report',                                                   # 54
      'Missing new file name for !0!.',                                  # 55
      'Missing directory name.',                                         # 56
      'Illegal Averist authentication method specified',                 # 57
      'Describe files and directories in !0!',                           # 58
      'Help',                                                            # 59
      'Illegal help topic specified',                                    # 60
      'Bad base directory specified in configuration',                   # 61
      'File size of target does not match source.',                      # 62
      'Would cause an infinite loop (deep recursion).',                  # 63
      'Copy of file !0! to !1! successful.',                             # 64
      'Copy of file !0! to !1! failed.',                                 # 65
      'Copy report',                                                     # 66
      'Couldn\'t read from source file.',                                # 67
      'Couldn\'t write to target file.',                                 # 68
      'Name',                                                            # 69
      'Size',                                                            # 70
      'Time',                                                            # 71
      'Rename directory !0! to !1!.',                                    # 72
      'Rename file !0! to !1!.',                                         # 73
      'Edit files in !0!',                                               # 74
      '!0! is not a plain file.',                                        # 75
      '!0! is not a text file.',                                         # 76
      'Couldn\'t open !0!.',                                             # 77
      'Edit !0!:',                                                       # 78
      'Save',                                                            # 79
      'Edit report',                                                     # 80
      'Save of file !0! successful.',                                    # 81
      'Save of file !0! failed.',                                        # 82
      'File size of saved !0! does not match buffer size.',              # 83
      'Filter',                                                          # 84
      'Filter report',                                                   # 85
      'Filter !0! on directory !1! successful.',                         # 86
      'Filter !0! on file !1! successful.',                              # 87
      'Filter !0! on directory !1! failed.',                             # 88
      'Filter !0! on file !1! failed.',                                  # 89
      'Filter !0! successful.',                                          # 90
      'Filter !0! failed.',                                              # 91
      'Filter !0! will create !1! with an appropriate extension.',       # 92
      'Illegal filter specified',                                        # 93
      '!0! is too large to edit.',                                       # 94
      'Describe',                                                        # 95
      'Bad character in directory name.',                                # 96
      'Bad character in file name.',                                     # 97
      'Bad character in new file name for !0!.',                         # 98
      'Login',                                                           # 99
      'Username:',                                                       # 100
      'Password:',                                                       # 101
      'Authentication needed or bad username specified.',                # 102
      'Reauthentication needed or bad password specified.',              # 103
      'Logout',                                                          # 104
      'You are now logged out.',                                         # 105
      'Secure (SSL) connection required',                                # 106
      'Authenticated username does not exist locally',                   # 107
      '(description)',                                                   # 108
      'Description saved.',                                              # 109
      'Couldn\'t save description.',                                     # 110
      'Describe directory !0!: !1!',                                     # 111
      'Describe file !0!: !1!',                                          # 112
      'Describe',                                                        # 113
      'Describe report',                                                 # 114
      'Descriptions saved.',                                             # 115
      'Couldn\'t save descriptions.',                                    # 116
      'Error',                                                           # 117
      'Target directory',                                                # 118
      'Continue',                                                        # 119
      'Bad directory specified. Try again.',                             # 120
      ],


## Setting this variable to "1" (true/yes) means that files only can be
## retrieved (downloaded) if a web server has access to them.
## This is the prefered setting if you only use the script as a "remote web
## server manager" because the security is higher. Also any file is sent with
## the MIME type set in the web server.
## Setting this variable to "0" (false/no) means that all retrieval will be
## handled by the script itself. This means that any file on the server
## readable by the script can be retrieved as long as it is not above the base
## directory.

     'only_true_web_server_retrieval' => 1,


## Base URL for the file manager. This setting is only used if the above
## setting is "1". If "1" is set above this URL must correspond to the base
## directory set further below. Anyway, it must have a value even if it's
## not used. If you use authentication you may use the line with the
## !username! variable which then is replaced with the authenticated user
## name.

     'base_url' => 'http://127.0.0.1/',
#    'base_url' => 'http://127.0.0.1/~!username!/',


## Base directory for the file manager. If you only allow file retrieval via
## the web server's access this must be set to the same directory as that the
## URL above correspond to. Otherwise you can set this to whatever directory
## you like.
## If you use authentication you may use the line with the !username! variable
## which then is replaced with the authenticated user name.
## The user will not be able to go above this directory.
## Be sure you know what you set this to, because if you configure this badly,
## like setting it too high up in your directory tree, you open up
## SECURITY RISKS. This option has a default value that should not work on
## your system as a security precaution. This is to force you to set this
## value yourself and not rely on the default one.

     'base_dir' => '/home/httpd/html/127.0.0.1/',
#    'base_dir' => '/home/!username!/public_html/',


## Enable or disable security check of symbolic links.
## This option should always be set to "1" unless you know what you are doing.
## Disabling it (setting this option to "0") means that the user will be
## allowed to follow symlinks that point above the configured base directory.
## This is a great SECURITY RISK if the user can create symbolic links on
## their own, for example through shell access. Drall does not allow symbolic
## link creation. LEAVE THIS AS "1" OR YOU WILL ENABLE A SECURITY HOLE.

     'symlink_security' => 1,


## Enable or disable the dual-frame interface.

     'frames' => 1,


## Enable or disable presentation of information in directory listing. 

     'list_checkbox'    => 1,
     'list_icon'        => 1,
     'list_name'        => 1, # Well, you have the option at least...
     'list_size'        => 1,
     'list_time'        => 1,
     'list_description' => 0,


## Enable or disable linking of information to actions.

     'action_icon' => 1, # Action is change directory or retrieve file
     'action_name' => 1, # Action is change directory or retrieve file


## Enable or disable each function and the functions' features

     'enable_edit'     => 1,
     'enable_copy'     => 1,
     'enable_rename'   => 1,
     'enable_move'     => 1,
     'enable_delete'   => 1,
     'enable_describe' => 0,
     'enable_upload'   => 1,
     'enable_mkdir'    => 1,
     'enable_filter'   => 0,
     'enable_help'     => 0,

     'edit_maximum_size'         => 60000,
     'enable_copy_recursively'   => 1,
     'enable_delete_recursively' => 1,
     'enable_upload_description' => 0,
     'return_to_maked_directory' => 1,


## Disable the directory listing of certain files and directories by
## specifying patterns to match as Perl regular expressions. The examples
## below are disabled by default.

     'hide_files_and_directories' =>
     [
#     '^\.',             # Hide files and directories starting with "."
#     '\.html$',         # Hide files and directories ending with ".html"
#     '^descript\.ion$', # Hide files and directories named "descript.ion" 
      ],


## Description data file name. The file format is backwards compatible with
## the famous descript.ion files used in for example 4DOS. Long file names
## (anything else than the old 8.3 standard) are put inside "", while 8.3 file
## names are not. Descriptions are separated from the file name with a space,
## and are limited to 511 characters. One file name and description per line.
## The description data file must be stored in the same directory as the files
## it describe.

     'description_filename' => 'descript.ion',


## Location of the external help documents. These are only used if you enable
## the help function. You have to write these yourself.

     'help_location_panel'    => 'http://127.0.0.1/drall/panel.html',
     'help_location_continue' => 'http://127.0.0.1/drall/continue.html',
     'help_location_edit'     => 'http://127.0.0.1/drall/edit.html',
     'help_location_rename'   => 'http://127.0.0.1/drall/rename.html',
     'help_location_describe' => 'http://127.0.0.1/drall/describe.html',
     'help_location_upload'   => 'http://127.0.0.1/drall/upload.html',
     'help_location_mkdir'    => 'http://127.0.0.1/drall/mkdir.html',


## This is support for a local frame on top of the two file manager frames
## where you can put local information as for example help links. Note that
## this only works if you have enabled frames above.
## Src is the URL to the html file to be displayed, size is the size in pixels
## or in percent of the total window height, resize specifies whether you want
## the local frame to be resizeable and scrolling specifies whether you want
## scolling devices on the frame border (answering "1" will display scrolling
## devices when necessary).

     'enable_local_frame'    => 0,
     'local_frame_src'       => 'http://127.0.0.1/drall/frame.html',
     'local_frame_size'      => '40',
     'local_frame_resize'    => 0,
     'local_frame_scrolling' => 0,


## The file extensions and corresponding MIME types you want this script to
## know about. These are used for getting the correct icon (if that is enabled
## above) and when you do file retrieval via the script.

     'ext2type' =>
     {
	 'hqx'     => 'application/mac-binhex40',
	 'mcd'     => 'application/mathcad',
	 'ma'      => 'application/mathematica',
	 'doc'     => 'application/msword',
	 'bin'     => 'application/octet-stream',
	 'exe'     => 'application/octet-stream',
	 'pdf'     => 'application/pdf',
	 'ai'      => 'application/postscript',
	 'eps'     => 'application/postscript',
	 'ps'      => 'application/postscript',
	 'rtf'     => 'application/rtf',
	 'fm'      => 'application/vnd.framemaker',
	 'frame'   => 'application/vnd.framemaker',
	 'frm'     => 'application/vnd.framemaker',
	 'mdb'     => 'application/vnd.ms-access',
	 'xls'     => 'application/vnd.ms-excel',
	 'ppt'     => 'application/vnd.ms-powerpoint',
	 'mpp'     => 'application/vnd.ms-project',
	 'wpd'     => 'application/wordperfect',
	 'Z'       => 'application/x-compress',
	 'csh'     => 'application/x-csh',
	 'dvi'     => 'application/x-dvi',
	 'gz'      => 'application/x-gzip',
	 'latex'   => 'application/x-latex',
	 'sh'      => 'application/x-sh',
	 'shar'    => 'application/x-shar',
	 'sea'     => 'application/x-stuffit',
	 'sit'     => 'application/x-stuffit',
	 'tar'     => 'application/x-tar',
	 'tcl'     => 'application/x-tcl',
	 'tex'     => 'application/x-tex',
	 'texinfo' => 'application/x-texinfo',
	 'texi'    => 'application/x-texinfo',
	 'zip'     => 'application/zip',
	 'aif'     => 'audio/aiff',
	 'aiff'    => 'audio/aiff',
	 'aifc'    => 'audio/aiff',
	 'au'      => 'audio/basic',
	 'snd'     => 'audio/basic',
	 'es'      => 'audio/echospeech',
	 'mid'     => 'audio/midi',
	 'midi'    => 'audio/midi',
	 'kar'     => 'audio/midi',
	 'mpga'    => 'audio/mpeg',
	 'mp2'     => 'audio/mpeg',
	 'mp3'     => 'audio/mpeg',
	 'vox'     => 'audio/voxware',
	 'ram'     => 'audio/x-pn-realaudio',
	 'rm'      => 'audio/x-pn-realaudio',
	 'rpm'     => 'audio/x-pn-realaudio-plugin',
	 'ra'      => 'audio/x-realaudio',
	 'wav'     => 'audio/x-wav',
	 'cgm'     => 'image/cgm',
	 'fif'     => 'image/fif',
	 'gif'     => 'image/gif',
	 'ief'     => 'image/ief',
	 'jpeg'    => 'image/jpeg',
	 'jpg'     => 'image/jpeg',
	 'jpe'     => 'image/jpeg',
	 'png'     => 'image/png',
	 'tiff'    => 'image/tiff',
	 'tif'     => 'image/tiff',
	 'wi'      => 'image/wavelet',
	 'ras'     => 'image/x-cmu-raster',
	 'cmx'     => 'image/x-cmx',
	 'dwg'     => 'image/x-dwg',
	 'dxf'     => 'image/x-dxf',
	 'dsf'     => 'image/x-mgx-dsf',
	 'bmp'     => 'image/x-ms-bmp',
	 'pict'    => 'image/x-pict',
	 'pnm'     => 'image/x-portable-anymap',
	 'pbm'     => 'image/x-portable-bitmap',
	 'pgm'     => 'image/x-portable-graymap',
	 'ppm'     => 'image/x-portable-pixmap',
	 'svf'     => 'image/x-svf',
	 'xbm'     => 'image/x-xbitmap',
	 'xpm'     => 'image/x-xpixmap',
	 'css'     => 'text/css',
	 'rtx'     => 'text/enriched',
	 'html'    => 'text/html',
	 'htm'     => 'text/html',
	 'asc'     => 'text/plain',
	 'c'       => 'text/plain',
	 'c++'     => 'text/plain',
	 'cc'      => 'text/plain',
	 'h'       => 'text/plain',
	 'pl'      => 'text/plain',
	 'txt'     => 'text/plain',
	 'sgml'    => 'text/sgml',
	 'sgm'     => 'text/sgml',
	 'tsv'     => 'text/tab-separated-values',
	 'talk'    => 'text/x-speech',
	 'xml'     => 'text/xml',
	 'mpeg'    => 'video/mpeg',
	 'mpg'     => 'video/mpeg',
	 'mpe'     => 'video/mpeg',
	 'qt'      => 'video/quicktime',
	 'mov'     => 'video/quicktime',
	 'vdo'     => 'video/vdo',
	 'viv'     => 'video/vivo',
	 'avi'     => 'video/x-msvideo',
	 'movie'   => 'video/x-sgi-movie',
	 },


## The URL to the icon you want associated with each "type".
## The icons used below are the ones included with the Apache web server.
## The first three are special. They specify "unknown" files (those you
## haven't listed in ext2type above), parent directories (..) and normal
## directories. If you don't have icon usage enabled (above) when you don't
## have to specify any icons. Then you just leave this section as it is.

     'type2icon' =>
     {
	 'x-drall/unknown'               => '/icons/unknown.gif',
	 'x-drall/parent'                => '/icons/back.gif',
	 'x-drall/directory'             => '/icons/folder.gif',
	 'application/mac-binhex40'      => '/icons/binhex.gif',
	 'application/mathcad'           => '/icons/image3.gif',
	 'application/mathematica'       => '/icons/image3.gif',
	 'application/msword'            => '/icons/layout.gif',
	 'application/octet-stream'      => '/icons/binary.gif',
	 'application/pdf'               => '/icons/layout.gif',
	 'application/postscript'        => '/icons/a.gif',
	 'application/rtf'               => '/icons/text.gif',
	 'application/vnd.framemaker'    => '/icons/layout.gif',
	 'application/vnd.ms-access'     => '/icons/box2.gif',
	 'application/vnd.ms-excel'      => '/icons/image3.gif',
	 'application/vnd.ms-powerpoint' => '/icons/image2.gif',
	 'application/vnd.ms-project'    => '/icons/box1.gif',
	 'application/wordperfect'       => '/icons/layout.gif',
	 'application/x-compress'        => '/icons/compressed.gif',
	 'application/x-csh'             => '/icons/script.gif',
	 'application/x-dvi'             => '/icons/dvi.gif',
	 'application/x-gzip'            => '/icons/compressed.gif',
	 'application/x-latex'           => '/icons/text.gif',
	 'application/x-sh'              => '/icons/script.gif',
	 'application/x-shar'            => '/icons/compressed.gif',
	 'application/x-stuffit'         => '/icons/compressed.gif',
	 'application/x-tar'             => '/icons/tar.gif',
	 'application/x-tcl'             => '/icons/script.gif',
	 'application/x-tex'             => '/icons/tex.gif',
	 'application/x-texinfo'         => '/icons/text.gif',
	 'application/zip'               => '/icons/compressed.gif',
	 'audio/aiff'                    => '/icons/sound1.gif',
	 'audio/basic'                   => '/icons/sound1.gif',
	 'audio/echospeech'              => '/icons/sound1.gif',
	 'audio/midi'                    => '/icons/sound2.gif',
	 'audio/mpeg'                    => '/icons/sound2.gif',
	 'audio/voxware'                 => '/icons/sound2.gif',
	 'audio/x-pn-realaudio'          => '/icons/sound1.gif',
	 'audio/x-pn-realaudio-plugin'   => '/icons/sound1.gif',
	 'audio/x-realaudio'             => '/icons/sound1.gif',
	 'audio/x-wav'                   => '/icons/sound1.gif',
	 'image/cgm'                     => '/icons/image1.gif',
	 'image/fif'                     => '/icons/image1.gif',
	 'image/gif'                     => '/icons/image1.gif',
	 'image/ief'                     => '/icons/image1.gif',
	 'image/jpeg'                    => '/icons/image1.gif',
	 'image/png'                     => '/icons/image1.gif',
	 'image/tiff'                    => '/icons/image1.gif',
	 'image/wavelet'                 => '/icons/image1.gif',
	 'image/x-cmu-raster'            => '/icons/image1.gif',
	 'image/x-cmx'                   => '/icons/image1.gif',
	 'image/x-dwg'                   => '/icons/image1.gif',
	 'image/x-dxf'                   => '/icons/image1.gif',
	 'image/x-mgx-dsf'               => '/icons/image1.gif',
	 'image/x-ms-bmp'                => '/icons/image1.gif',
	 'image/x-pict'                  => '/icons/image1.gif',
	 'image/x-portable-anymap'       => '/icons/image1.gif',
	 'image/x-portable-bitmap'       => '/icons/image1.gif',
	 'image/x-portable-graymap'      => '/icons/image1.gif',
	 'image/x-portable-pixmap'       => '/icons/image1.gif',
	 'image/x-svf'                   => '/icons/image1.gif',
	 'image/x-xbitmap'               => '/icons/image1.gif',
	 'image/x-xpixmap'               => '/icons/image1.gif',
	 'text/css'                      => '/icons/text.gif',
	 'text/enriched'                 => '/icons/text.gif',
	 'text/html'                     => '/icons/layout.gif',
	 'text/plain'                    => '/icons/text.gif',
	 'text/sgml'                     => '/icons/text.gif',
	 'text/tab-separated-values'     => '/icons/text.gif',
	 'text/x-speech'                 => '/icons/sound1.gif',
	 'text/xml'                      => '/icons/text.gif',
	 'video/mpeg'                    => '/icons/movie.gif',
	 'video/quicktime'               => '/icons/movie.gif',
	 'video/vdo'                     => '/icons/movie.gif',
	 'video/vivo'                    => '/icons/movie.gif',
	 'video/x-msvideo'               => '/icons/movie.gif',
	 'video/x-sgi-movie'             => '/icons/movie.gif',
     },


## What MIME type to use when we do file retrieval via the script and the file
## type is unknown. This must be a type that really sends binary data so that
## information isn't corrupted when transferred. The below is usually the
## best. Leave this as it is if you don't know what this is.

     'script_retrieval_unknown_mime_type' => 'application/octet-stream',


## Setting this variable to "1" (true/yes) means that any file retrieved via
## the script is sent with the MIME type for files of unknown type. This
## "emulates" FTP client behaviour.
## Setting this variable to "0" (false/no) means that if there is a MIME type
## given for the file it will be sent with that, otherwise the MIME type for
## files of unknown type will be used.
## If you want the file manager to behave exactly like an FTP client set this
## to "1". To be able to view files, like web pages and images, directly in
## the web browser while using script retrieval you need to leave this as "0".

     'script_retrieval_always_unknown' => 0,


## Filter allows the user to perform some sort of action on the selected
## files. Each filter has two variables, description and command. The
## description is what the user will see in the drop down menu if filters are
## enabled, and command is what action the script will perform. There are
## certain variables that can be used in the command.
## !s! is the name of a single file. If this is used in the command, the
## command will be performed once, with this variable replaced with the
## file name, for each selected file.
## !l! is a space separated list of the selected files. If this is used in the
## command, the command will be performed only once for all the selected
## files.
## !t! is the true directory where the selected files are located.
## !o! will be replaced with an unique string that should be used for an
## output file name.
## The command must contain at least one of the !s! and !l! variables or
## nothing will be performed. Before the command is performed the working
## directory is changed to !t!. See given (disabled) examples below for some
## tips.

     'filter' =>
     [
#     {'description' => 'zip',
#      'command'     => 'zip -r !o!.zip !l!'},
#     {'description' => 'unzip',
#      'command'     => 'unzip -n !s!'},
#     {'description' => 'tar + gzip',
#      'command'     => 'tar -c -f !o!.tar.gz -C !t! -z !l!'},
#     {'description' => 'gunzip + untar',
#      'command'     => 'tar -x -f !s! -z'},
#     {'description' => 'tar + compress',
#      'command'     => 'tar -c -f !o!.tar.Z -C !t! -Z !l!'},
#     {'description' => 'uncompress + untar',
#      'command'     => 'tar -x -f !s! -Z'},
      ],
     
     
## UNIX rights for new files/directories. UNIX people should know what this
## means. UNIX admins should make sure these are set correctly.

     'filemode' => 0644,
     'dirmode'  => 0755,
	   

## Some internal MIME types. Please leave these alone... :-)

     'type_unknown'   => 'x-drall/unknown',
     'type_parent'    => 'x-drall/parent',
     'type_directory' => 'x-drall/directory',
     'type_hidden'    => 'x-drall/hidden',
     'type_disallow'  => 'x-drall/disallow',


## The title for the file manager.
## Set this to what you want on the browser's title row.

     'title_default'       => 'Drall',
     'title_authenticated' => 'Drall for !username!',


## Colors. The colors the script will use on the HTML page. If you want
## standard browser colors just set these to '' (empty quote).

     'color_background' => '#ffffff',
     'color_text'       => '#000000',
     'color_link'       => '#0000ff',


## Enable or disable the display of a link to the home page of Drall (and to
## the home page of Averist if authentication is enabled) at the bottom of
## each page.
## Remember that you were given this product for free so the least you can do
## is to display one or two links. Please leave this as "1".

     'advertise' => 1,


## The path to the script. You can leave this alone. Please do... :-)

     'script_name' => $ENV{'SCRIPT_NAME'},


     );


######################
# Configuration Ends #
######################



########  Here starts the code section. Do NOT edit anything from here on if
# Main #  you don't know what you are doing. If you are planning to tweak then
########  please read the License first.


use vars qw(%query);
use vars qw($username $ticket);
use vars qw($file);
use vars qw(%directory $panel $opanel);
if (defined($config{'config_directory'})) {
    $config{'config_directory'} = fix_path("$config{'config_directory'}\/");
}
if (defined($config{'base_dir'})) {
    $config{'base_dir'} = fix_path("$config{'base_dir'}\/");
}
if (defined($config{'base_url'})) {
    $config{'base_url'} = fix_path("$config{'base_url'}\/");
}
if (! defined($config{'script_name'})) {
    $config{'script_name'} = '';
}
%query = parse_query();
if ((fix_config()) && (check_ssl()) && (authenticate()) && (suid())) {
    if (defined ($username)) {
	if ($config{'config_user'}) {
	    $file = "$config{'config_directory'}$username.user.drall.pl";
	    if (-r $file) {
		require $file;
	    }
	}
    }
    if (defined ($username)) {
	substitute_recursive(\%config,"username",$username);
    }
    if ((! defined($query{'left'})) || ($query{'left'} eq '')) {
	$directory{'left'} = '/';
    }
    else {
	$directory{'left'} = fix_path("\/$query{'left'}\/");
    }
    if ((! defined($query{'right'})) || ($query{'right'} eq '')) {
	$directory{'right'} = '/';
    }
    else {
	$directory{'right'} = fix_path("\/$query{'right'}\/");
    }
    if (defined($query{'panel'})) {
	$panel = $query{'panel'};
    }
    else {
	$panel = 'left';
    }
    if ($panel eq 'left') {
	$opanel = 'right';
    }
    else {
	$opanel = 'left';
    }
    if ((! defined($config{'base_dir'})) ||
	(! validate_directory($config{'base_dir'},''))) {
	do_error($config{'language'}[61]);
    }
    elsif ((! $config{'frames'}) &&
	   (defined($query{'continue'})) &&
	   (! validate_directory($config{'base_dir'},$directory{$opanel}))) {
	get_target_directory($config{'language'}[120]);
    }
    else {
	if (! validate_directory($config{'base_dir'},$directory{'left'})) {
	    if (validate_directory($config{'base_dir'},$directory{'right'})) {
		$directory{'left'} = $directory{'right'};
	    }
	    else {
		$directory{'left'} = '/';
		$directory{'right'} = '/';
	    }
	}
	else {
	    if (! validate_directory($config{'base_dir'},
				     $directory{'right'})) {
		$directory{'right'} = $directory{'left'};
	    }
	}
	if (! defined($query{'action'})) {
	    $query{'action'} = 'panels';
	}
	if (defined($query{'edit'})) {
	    if ($config{'enable_edit'}) {
		get_edit();
	    }
	    else {
		do_error($config{'language'}[42]);
	    }
	}
	elsif (defined($query{'copy'})) {
	    if ($config{'enable_copy'}) {
		if (($config{'frames'}) || (defined($query{'continue'}))) {
		    do_copy();
		}
		else {
		    get_target_directory();
		}
	    }
	    else {
		do_error($config{'language'}[42]);
	    }
	}
	elsif (defined($query{'rename'})) {
	    if ($config{'enable_rename'}) {
		get_rename();
	    }
	    else {
		do_error($config{'language'}[42]);
	    }
	}
	elsif (defined($query{'move'})) {
	    if ($config{'enable_move'}) {
		if (($config{'frames'}) || (defined($query{'continue'}))) {
		    do_move();
		}
		else {
		    get_target_directory();
		}
	    }
	    else {
		do_error($config{'language'}[42]);
	    }
	}
	elsif (defined($query{'delete'})) {
	    if ($config{'enable_delete'}) {
		do_delete();
	    }
	    else {
		do_error($config{'language'}[42]);
	    }
	}
	elsif (defined($query{'describe'})) {
	    if ($config{'enable_describe'}) {
		get_describe();
	    }
	    else {
		do_error($config{'language'}[42]);
	    }
	}
	elsif (defined($query{'filter'})) {
	    if ($config{'enable_filter'}) {
		if (defined($config{'filter'}[$query{'filtern'}])) {
		    do_filter();
		}
		else {
		    do_error($config{'language'}[93]);
		}
	    }
	    else {
		do_error($config{'language'}[42]);
	    }
	}
	elsif (defined($query{'upload'})) {
	    if ($config{'enable_upload'}) {
		get_upload(0);
	    }
	    else {
		do_error($config{'language'}[42]);
	    }
	}
	elsif (defined($query{'mkdir'})) {
	    if ($config{'enable_mkdir'}) {
		get_mkdir(0);
	    }
	    else {
		do_error($config{'language'}[42]);
	    }
	}
	elsif (defined($query{'logout'})) {
	    if ($config{'authentication'}) {
		Averist::logout();
		undef $username;
		undef $ticket;
		delete $query{'logout'};
		print_login($config{'language'}[105]);
	    }
	    else {
		do_error($config{'language'}[42]);
	    }
	}
	elsif (defined($query{'help'})) {
	    if ($config{'enable_help'}) {
		if (defined($query{'topic'})) {
		    if (defined($config{"help_location_$query{'topic'}"})) {
			print "Location: ";
			print $config{"help_location_$query{'topic'}"};
			print "\n\n";
		    }
		    else {
			do_error($config{'language'}[60]);
		    }
		}
		else {
		    do_error($config{'language'}[60]);
		}
	    }
	    else {
		do_error($config{'language'}[42]);
	    }
	}
	elsif ($query{'action'} eq 'panels') {
	    if ($config{'frames'}) {
		print_frames();
	    }
	    else {
		print_directory_listing();
	    }
	}   
	elsif ($query{'action'} eq 'list') {
	    print_directory_listing();
	}
	elsif ($query{'action'} eq 'retrieve') {
	    if (defined($query{'directory'})) {
		if (defined($query{'file'})) {
		    retrieve();
		}
		else {
		    do_error($config{'language'}[1]);
		}
	    }
	    else {
		do_error($config{'language'}[0]);
	    }
	}   
	elsif ($query{'action'} eq 'edit') {
	    if ($config{'enable_edit'}) {
		do_edit();
	    }
	    else {
		do_error($config{'language'}[42]);
	    }
	}
	elsif ($query{'action'} eq 'rename') {
	    if ($config{'enable_rename'}) {
		do_rename();
	    }
	    else {
		do_error($config{'language'}[42]);
	    }
	}
	elsif ($query{'action'} eq 'describe') {
	    if ($config{'enable_describe'}) {
		do_describe();
	    }
	    else {
		do_error($config{'language'}[42]);
	    }
	}
	elsif ($query{'action'} eq 'upload') {
	    if ($config{'enable_upload'}) {
		do_upload();
	    }
	    else {
		do_error($config{'language'}[42]);
	    }
	}
	elsif ($query{'action'} eq 'mkdir') {
	    if ($config{'enable_mkdir'}) {
		do_mkdir();
	    }
	    else {
		do_error($config{'language'}[42]);
	    }
	}
	else {
	    do_error($config{'language'}[2]);
	}
    }
}


#############
# Main Ends #
#############



#############  Code section continues. Here with the functions.
# Functions #
#############


#######################  Functions used for interfacing with the client.
# Interface Functions #
#######################

sub print_head {
    print "Content-Type: text/html\n\n";
    print "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" ";
    print "\"http://www.w3.org/TR/html401/loose.dtd\">\n";
    print "<HTML>\n";
    print "<HEAD>\n";
    print "<TITLE>";
    if (defined($ticket)) {
	print $config{'title_authenticated'};
    }
    else {
	print $config{'title_default'};
    }
    print "</TITLE>\n";
    print "<META NAME=\"generator\" CONTENT=\"$product_name $product_version";
    if (($config{'authentication'}) &&
	(defined($Averist::Product_Name)) &&
	(defined($Averist::Product_Version))) {
	print " / $Averist::Product_Name $Averist::Product_Version";
    }
    print "\">\n";
    print "</HEAD>\n";
    print "<BODY";
    print " BGCOLOR=\"$config{'color_background'}\"";
    print " TEXT=\"$config{'color_text'}\"";
    print " LINK=\"$config{'color_link'}\"";
    print " VLINK=\"$config{'color_link'}\"";
    print " ALINK=\"$config{'color_link'}\"";
    print ">\n";
}

sub print_foot {
    if ($config{'advertise'}) {
	print "<P ALIGN=\"right\"><EM>";
	print "<A HREF=\"http://www.edlund.org/hacks/drall/\" ";
	print "TARGET=\"_blank\">$product_name $product_version</A>";
	if (($config{'authentication'}) &&
	    (defined($Averist::Product_Name)) &&
	    (defined($Averist::Product_Version))) {
	    print " / ";
	    print "<A HREF=\"http://www.edlund.org/hacks/averist/\" ";
	    print "TARGET=\"_blank\">$Averist::Product_Name ";
	    print "$Averist::Product_Version</A>";
	}
	print "</EM></P>\n";
    }
    print "</BODY>\n";
    print "</HTML>\n";
}

sub print_login {
    print_head();
    print "<H1>$config{'language'}[99]</H1>\n";
    print "<P><FORM ACTION=\"$config{'script_name'}\" METHOD=\"post\" ";
    print "TARGET=\"_self\">\n";
    foreach (keys %query) {
	if (($_ ne $Averist::Config{'methods'}{'authentication'}{'remote'}
	     {'cgi'}{'username'}) &&
	    ($_ ne $Averist::Config{'methods'}{'authentication'}{'remote'}
	     {'cgi'}{'passphrase'}) &&
	    ($_ ne $Averist::Config{'methods'}{'reauthentication'}{'remote'}
	     {'cgi'}{'username'}) &&
	    ($_ ne $Averist::Config{'methods'}{'reauthentication'}{'remote'}
	     {'cgi'}{'ticket'})) {
	    print "<INPUT TYPE=\"hidden\" NAME=\"encoded_$_\" VALUE=\"";
	    print uri_escape_everything($query{$_});
	    print "\">\n";
	}
    }
    if (defined ($_[0])) {
	print "<P>$_[0]\n";
    }
    elsif (! defined($username)) {
	print "<P>$config{'language'}[102]\n";
    }
    elsif (! defined($ticket)) {
	print "<P>$config{'language'}[103]\n";
    }
    print "<P>$config{'language'}[100]<BR>\n";
    print "<INPUT TYPE=\"text\" NAME=\"";
    print $Averist::Config{'methods'}{'authentication'}{'remote'}{'cgi'}
    {'username'};
    print "\" VALUE=\"";
    if (defined($username)) {
	print character_reference($username);
    }
    print "\" SIZE=\"32\" MAXLENGTH=\"32\">\n";
    print "<P>$config{'language'}[101]<BR>\n";
    print "<INPUT TYPE=\"password\" NAME=\"";
    print $Averist::Config{'methods'}{'authentication'}{'remote'}{'cgi'}
    {'passphrase'};
    print "\" VALUE=\"\" SIZE=\"32\" MAXLENGTH=\"128\">\n";
    print "<P><INPUT TYPE=\"submit\" ";
    print "VALUE=\"$config{'language'}[99]\">\n";
    print "</FORM>\n";
    print_foot();
}

sub print_cgi_reauth {
    if (($config{'authentication'}) &&
	($Averist::Config{'events'}{'reauthentication'}{'remote'} eq "cgi")) {
	print "<INPUT TYPE=\"hidden\" NAME=\"";
	print $Averist::Config{'methods'}{'reauthentication'}{'remote'}{'cgi'}
	{'username'};
	print "\" VALUE=\"";
	print character_reference($username);
	print "\">\n";
	print "<INPUT TYPE=\"hidden\" NAME=\"";
	print $Averist::Config{'methods'}{'reauthentication'}{'remote'}{'cgi'}
	{'ticket'};
	print "\" VALUE=\"";
	print character_reference($ticket);
	print "\">\n";
    }
}

sub print_config_and_order {
    if (defined($query{'config'})) {
	print "<INPUT TYPE=\"hidden\" NAME=\"config\" VALUE=\"";
	print character_reference($query{'config'});
	print "\">\n";
    }
    if (defined($query{"order-$panel"})) {
	print "<INPUT TYPE=\"hidden\" NAME=\"order-$panel\" VALUE=\"";
	print $query{"order-$panel"};
	print "\">\n";
    }
    if (defined($query{"order-$opanel"})) {
	print "<INPUT TYPE=\"hidden\" NAME=\"order-$opanel\" VALUE=\"";
	print $query{"order-$opanel"};
	print "\">\n";
    }
}

sub print_panel_and_directory {
    print "<INPUT TYPE=\"hidden\" NAME=\"panel\" VALUE=\"$panel\">\n";
    print "<INPUT TYPE=\"hidden\" NAME=\"left\" VALUE=\"";
    print character_reference($directory{'left'});
    print "\">\n";
    print "<INPUT TYPE=\"hidden\" NAME=\"right\" VALUE=\"";
    print character_reference($directory{'right'});
    print "\">\n";
}

sub print_action_return {
    my $help = $_[0];
    print "<P><HR>\n";
    print "<P><FORM ACTION=\"$config{'script_name'}\" METHOD=\"post\" ";
    print "TARGET=\"_self\">\n";
    print_cgi_reauth();
    if (defined($query{'config'})) {
	print "<INPUT TYPE=\"hidden\" NAME=\"config\" ";
	print "VALUE=\"$query{'config'}\">\n";
    }
    if (defined($query{"order-$panel"})) {
	print "<INPUT TYPE=\"hidden\" NAME=\"order-$panel\" VALUE=\"";
	print $query{"order-$panel"};
	print "\">\n";
    }
    if (defined($query{"order-$opanel"})) {
	print "<INPUT TYPE=\"hidden\" NAME=\"order-$opanel\" VALUE=\"";
	print $query{"order-$opanel"};
	print "\">\n";
    }
    print "<INPUT TYPE=\"hidden\" NAME=\"action\" VALUE=\"panels\">\n";
    print "<INPUT TYPE=\"hidden\" NAME=\"left\" VALUE=\"";
    print character_reference($directory{'left'});
    print "\">\n";
    print "<INPUT TYPE=\"hidden\" NAME=\"right\" VALUE=\"";
    print character_reference($directory{'right'});
    print "\">\n";
    print "<INPUT TYPE=\"submit\" VALUE=\"$config{'language'}[18]\">\n";
    if (($config{'enable_help'}) && ($help ne "")) {
	print "<BR>\n";
	print "<INPUT TYPE=\"hidden\" NAME=\"topic\" VALUE=\"$help\">";
	print "<INPUT TYPE=\"submit\" NAME=\"help\" ";
	print "VALUE=\"$config{'language'}[59]\">\n";
    }
    print "</FORM>\n";
}

sub print_frames {
    my $url;
    my $lurl;
    my $rurl;
    $url = "$config{'script_name'}?";
    if (($config{'authentication'}) &&
	($Averist::Config{'events'}{'reauthentication'}{'remote'} eq "cgi")) {
	if (defined($username)) {
	    $url .= $Averist::Config{'methods'}{'reauthentication'}{'remote'}
	    {'cgi'}{'username'};
	    $url .= "\=" . uri_escape($username) . "\;";
	}
	if (defined($ticket)) {
	    $url .= $Averist::Config{'methods'}{'reauthentication'}{'remote'}
	    {'cgi'}{'ticket'};
	    $url .= "\=" . uri_escape($ticket) . "\;";
	}
    }
    if (defined($query{'config'})) {
	$url .= "config=$query{'config'}\;";
    }
    $url .= "action=list";
    $lurl = $url;
    $lurl .= ";panel=left";
    $rurl = $url;
    $rurl .= ";panel=right";
    $url = ";left=";
    $url .= uri_escape($directory{'left'});
    $url .= ";right=";
    $url .= uri_escape($directory{'right'});
    if (defined($query{'order-left'})) {
	$url .= ";order-left=$query{'order-left'}";
    }
    if (defined($query{'order-right'})) {
	$url .= ";order-right=$query{'order-right'}";
    }
    $lurl .= $url;
    $rurl .= $url;
    print "Content-Type: text/html\n\n";
    print "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Frameset//EN\" ";
    print "\"http://www.w3.org/TR/html401/frameset.dtd\">\n";
    print "<HTML>\n";
    print "<HEAD>\n";
    print "<TITLE>";
    if (defined($ticket)) {
	print $config{'title_authenticated'};
    }
    else {
	print $config{'title_default'};
    }
    print "</TITLE>\n";
    print "<META NAME=\"generator\" CONTENT=\"$product_name $product_version";
    if (($config{'authentication'}) &&
	(defined($Averist::Product_Name)) &&
	(defined($Averist::Product_Version))) {
	print " / $Averist::Product_Name $Averist::Product_Version";
    }
    print "\">\n";
    print "</HEAD>\n";
    if ($config{'enable_local_frame'}) {
	print "<FRAMESET ROWS=\"$config{'local_frame_size'}, *\">\n";
	print "<FRAME SRC=\"$config{'local_frame_src'}\"";
	if ($config{'local_frame_resize'} == 0) {
	    print " NORESIZE";
	}
	if ($config{'local_frame_scrolling'} == 0) {
	    print " SCROLLING=\"no\"";
	}
	print ">\n";
    }
    print "<FRAMESET COLS=\"50%, 50%\">\n";
    print "<FRAME SRC=\"$lurl\">\n";
    print "<FRAME SRC=\"$rurl\">\n";
    print "</FRAMESET>\n";
    if ($config{'enable_local_frame'}) {
	print "</FRAMESET>\n";
    }
    print "</HTML>\n"; 
}

sub print_directory_listing {
    my $url;
    my $file;
    my @dirs;
    my $i;
    my $descriptions;
    my @localtime;
    print_head();
    print "<FORM ACTION=\"$config{'script_name'}\" METHOD=\"post\" ";
    print "TARGET=\"_parent\">\n";
    print_cgi_reauth();
    print_config_and_order();
    print_panel_and_directory();
    print "<TABLE CELLPADDING=\"3\" CELLSPACING=\"0\">\n";
    print "<TR>\n";
    print "<TD COLSPAN=\"";
    print $config{'list_checkbox'}+
	$config{'list_icon'}+
	    $config{'list_name'}+
		$config{'list_size'}+
		    ($config{'list_time'}*2);
    print "\">";
    substitute(\$config{'language'}[3],
	       character_reference($directory{$panel}));
    print "<H1>$config{'language'}[3]</H1></TD>\n";
    print "</TR>\n";
    print "<TR>\n";
    if ($config{'list_checkbox'} == 1) {
	print "<TD></TD>\n";
    }
    if ($config{'list_icon'} == 1) {
	print "<TD></TD>\n";
    }
    $url = "$config{'script_name'}?";
    if (($config{'authentication'}) &&
	($Averist::Config{'events'}{'reauthentication'}{'remote'} eq "cgi")) {
	if (defined($username)) {
	    $url .= $Averist::Config{'methods'}{'reauthentication'}{'remote'}
	    {'cgi'}{'username'};
	    $url .= "\=" . uri_escape($username) . "\;";
	}
	if (defined($ticket)) {
	    $url .= $Averist::Config{'methods'}{'reauthentication'}{'remote'}
	    {'cgi'}{'ticket'};
	    $url .= "\=" . uri_escape($ticket) . "\;";
	}
    }
    if (defined($query{'config'})) {
	$url .= "config=$query{'config'}\;";
    }
    $url .= "action=panels";
    $url .= ";left=";
    $url .= uri_escape($directory{'left'});
    $url .= ";right=";
    $url .= uri_escape($directory{'right'});
    if (! defined($query{"order-$panel"})) {
	$query{"order-$panel"} = 'name';
    }
    if (! defined($query{"order-$opanel"})) {
	$query{"order-$opanel"} = 'name';
    }
    if ($config{'list_name'} == 1) {
	print "<TD>";
	print "<STRONG><A HREF=\"$url";
	if ($query{"order-$panel"} eq 'name') {
	    print ";order-$panel=name-reverse";
	}
	else {
	    print ";order-$panel=name";
	}
	print ";order-$opanel=";
	print $query{"order-$opanel"};
	print "\" TARGET=\"_parent\">$config{'language'}[69]</A></STRONG>";
	print "</TD>\n";
    }
    if ($config{'list_size'} == 1) {
	print "<TD ALIGN=\"right\">";
	print "<STRONG><A HREF=\"$url";
	if ($query{"order-$panel"} eq 'size') {
	    print ";order-$panel=size-reverse";
	}
	else {
	    print ";order-$panel=size";
	}
	print ";order-$opanel=";
	print $query{"order-$opanel"};
	print "\" TARGET=\"_parent\">$config{'language'}[70]</A></STRONG>";
	print "</TD>\n";
    }
    if ($config{'list_time'} == 1) {
	print "<TD ALIGN=\"center\" COLSPAN=\"2\">";
	print "<STRONG><A HREF=\"$url";
	if ($query{"order-$panel"} eq 'time') {
	    print ";order-$panel=time-reverse";
	}
	else {
	    print ";order-$panel=time";
	}
	print ";order-$opanel=";
	print $query{"order-$opanel"};
	print "\" TARGET=\"_parent\">$config{'language'}[71]</A></STRONG>";
	print "</TD>\n";
    }
    print "</TR>\n";
    if ($directory{$panel} ne "/") {
	print "<TR>\n";
	$url = "$config{'script_name'}?";
	if (($config{'authentication'}) &&
	    ($Averist::Config{'events'}{'reauthentication'}{'remote'}
	     eq "cgi")) {
	    if (defined($username)) {
		$url .= $Averist::Config{'methods'}{'reauthentication'}
		{'remote'}{'cgi'}{'username'};
		$url .= "\=" . uri_escape($username) . "\;";
	    }
	    if (defined($ticket)) {
		$url .= $Averist::Config{'methods'}{'reauthentication'}
		{'remote'}{'cgi'}{'ticket'};
		$url .= "\=" . uri_escape($ticket) . "\;";
	    }
	}
	if (defined($query{'config'})) {
	    $url .= "config=$query{'config'}\;";
	}
	$url .= "action=panels";
	$file = '';
	@dirs = split(/\//,$directory{$panel});
	for ($i = 1; $i < $#dirs; $i++) {
	    $file .= "/$dirs[$i]";
	}
	if ($panel eq "left") {
	    $url .= ";left=";
	    $url .= uri_escape($file);
	    $url .= "/;right=";
	    $url .= uri_escape($directory{'right'});
	}
	if ($panel eq "right") {
	    $url .= ";left=";
	    $url .= uri_escape($directory{'left'});
	    $url .= ";right=";
	    $url .= uri_escape($file);
	    $url .= "/";
	}
	$url .= ";order-$panel=";
	$url .= $query{"order-$panel"};
	$url .= ";order-$opanel=";
	$url .= $query{"order-$opanel"};
	if ($config{'list_checkbox'} == 1) {
	    print "<TD></TD>\n";
	}
	if ($config{'list_icon'} == 1) {
	    print "<TD>";
	    if ($config{'action_icon'} == 1) {
		print "<A HREF=\"$url\" TARGET=\"_parent\">";
	    }
	    print "<IMG SRC=\"$config{'type2icon'}{$config{'type_parent'}}\"";
	    print "ALT=\"\" BORDER=\"0\">";
	    if ($config{'action_icon'} == 1) {
		print "</A>";
	    }
	    print "</TD>\n";
	}
	if ($config{'list_name'} == 1) {
	    print "<TD>";
	    if ($config{'action_name'} == 1) {
		print "<A HREF=\"$url\" TARGET=\"_parent\">";
	    }
	    print "$config{'language'}[4]";
	    if ($config{'action_name'} == 1) {
		print "</A>";
	    }
	    print "</TD>\n";
	}
	if ($config{'list_size'} == 1) {
	    print "<TD></TD>\n";
	}
	if ($config{'list_time'} == 1) {
	    print "<TD></TD>\n";
	}
	print "</TR>\n";
    }
    if ($config{'list_description'}) {
	$descriptions = read_descriptions($directory{$panel});
    }
    foreach $file (@{read_directory($directory{$panel})}) {
	if (($$file{'type'} ne $config{'type_hidden'}) &&
	    ($$file{'type'} ne $config{'type_disallow'})) {
	    print "<TR>\n";
	    if ($$file{'type'} eq $config{'type_directory'}) {
		$url = "$config{'script_name'}?";
		if (($config{'authentication'}) &&
		    ($Averist::Config{'events'}{'reauthentication'}{'remote'}
		     eq "cgi")) {
		    if (defined($username)) {
			$url .= $Averist::Config{'methods'}
			{'reauthentication'}{'remote'}{'cgi'}{'username'};
			$url .= "\=" . uri_escape($username) . "\;";
		    }
		    if (defined($ticket)) {
			$url .= $Averist::Config{'methods'}
			{'reauthentication'}{'remote'}{'cgi'}{'ticket'};
			$url .= "\=" . uri_escape($ticket) . "\;";
		    }
		}
		if (defined($query{'config'})) {
		    $url .= "config=$query{'config'}\;";
		}
		$url .= "action=panels";
		if ($panel eq "left") {
		    $url .= ";left=";
		    $url .= uri_escape("$directory{'left'}$$file{'name'}");
		    $url .= "/;right=";
		    $url .= uri_escape($directory{'right'});
		}
		if ($panel eq "right") {
		    $url .= ";left=";
		    $url .= uri_escape($directory{'left'});
		    $url .= ";right=";
		    $url .= uri_escape("$directory{'right'}$$file{'name'}");
		    $url .= "/";
		}
		$url .= ";order-$panel=";
		$url .= $query{"order-$panel"};
		$url .= ";order-$opanel=";
		$url .= $query{"order-$opanel"};
	    }
	    else {
		if ($config{'only_true_web_server_retrieval'} == 1) {
		    $url = $$file{'url'};
		}
		else {
		    $url = $config{'script_name'};
		    $url .= "/";
		    $url .= underscore($$file{'name'});
		    $url .= "?";
		    if (($config{'authentication'}) &&
			($Averist::Config{'events'}{'reauthentication'}
			 {'remote'} eq "cgi")) {
			if (defined($username)) {
			    $url .= $Averist::Config{'methods'}
			    {'reauthentication'}{'remote'}{'cgi'}{'username'};
			    $url .= "\=" . uri_escape($username) . "\;";
			}
			if (defined($ticket)) {
			    $url .= $Averist::Config{'methods'}
			    {'reauthentication'}{'remote'}{'cgi'}{'ticket'};
			    $url .= "\=" . uri_escape($ticket) . "\;";
			}
		    }
		    if (defined($query{'config'})) {
			$url .= "config=$query{'config'}\;";
		    }
		    $url .= "action=retrieve;directory=";
		    $url .= uri_escape($directory{$panel});
		    $url .= ";file=";
		    $url .= uri_escape($$file{'name'});
		    $url .= ";filename=/";
		    $url .= $$file{'name'};
		}
	    }
	    if ($config{'list_checkbox'} == 1) {
		print "<TD><INPUT TYPE=\"checkbox\" ";
		print "NAME=\"(checked)$$file{'htmlname'}\"></TD>\n";
	    }
	    if ($config{'list_icon'} == 1) {
		print "<TD>";
		if ($config{'action_icon'} == 1) {
		    print "<A HREF=\"$url\" TARGET=\"_parent\">";
		}
		print "<IMG SRC=\"$config{'type2icon'}{$$file{'type'}}\" ";
		print "ALT=\"\" BORDER=\"0\">";
		if ($config{'action_icon'} == 1) {
		    print "</A>";
		}
		print "</TD>\n";
	    }
	    if ($config{'list_name'} == 1) {
		print "<TD>";
		if ($config{'action_name'} == 1) {
		    print "<A HREF=\"$url\" TARGET=\"_parent\">";
		}
		print $$file{'htmlname'};
		if ($config{'action_name'} == 1) {
		    print "</A>";
		}
		print "</TD>\n";
	    }
	    if ($config{'list_size'} == 1) {
		if ($$file{'type'} eq $config{'type_directory'}) {
		    print "<TD></TD>\n";
		}
		else {
		    print "<TD ALIGN=\"right\">$$file{'size'}</TD>\n";
		}
	    }
	    if ($config{'list_time'} == 1) {
		@localtime = localtime($$file{'time'});
		print "<TD ALIGN=\"center\">";
		print sprintf("%u-%02u-%02u",
			      $localtime[5]+1900,
			      $localtime[4]+1,
			      $localtime[3]);
		print "</TD>\n";
		print "<TD ALIGN=\"center\">";
		print sprintf("%02u:%02u:%02u",
			      $localtime[2],
			      $localtime[1],
			      $localtime[0]);
		print "</TD>\n";
	    }
	    print "</TR>\n";
	    if (($config{'list_description'}) &&
		(defined($$descriptions{$$file{'name'}}))) {
		print "<TR>\n";
		print "<TD COLSPAN=\"";
		print $config{'list_checkbox'}+
		    $config{'list_icon'}+
			$config{'list_name'}+
			    $config{'list_size'}+
				($config{'list_time'}*2);
		print "\">";
		print character_reference($$descriptions{$$file{'name'}});
		print "</TD>\n";
		print "</TR>\n";
	    }
	}
    }
    print "</TABLE>\n";
    print "<P><HR>\n<P>";
    # First line of buttons
    $i = 0;
    if ($config{'enable_edit'}) {
	print "<INPUT TYPE=\"submit\" NAME=\"edit\" ";
	print "VALUE=\"$config{'language'}[5]\">";
	$i = 1;
    }
    if ($config{'enable_copy'}) {
	if ($i == 1) {
	    print " ";
	}
	print "<INPUT TYPE=\"submit\" NAME=\"copy\" ";
	print "VALUE=\"$config{'language'}[6]\">";
	$i = 1;
    }
    if ($config{'enable_rename'}) {
	if ($i == 1) {
	    print " ";
	}
	print "<INPUT TYPE=\"submit\" NAME=\"rename\" ";
	print "VALUE=\"$config{'language'}[7]\">";
	$i = 1;
    }
    if ($config{'enable_move'}) {
	if ($i == 1) {
	    print " ";
	}
	print " <INPUT TYPE=\"submit\" NAME=\"move\" ";
	print "VALUE=\"$config{'language'}[8]\">";
	$i = 1;
    }
    if ($config{'enable_delete'}) {
	if ($i == 1) {
	    print " ";
	}
	print "<INPUT TYPE=\"submit\" NAME=\"delete\" ";
	print "VALUE=\"$config{'language'}[9]\">";
	$i = 1;
    }
    if ($config{'enable_describe'}) {
	if ($i == 1) {
	    print " ";
	}
	print "<INPUT TYPE=\"submit\" NAME=\"describe\" ";
	print "VALUE=\"$config{'language'}[95]\">";
	$i = 1;
    }
    if ($i == 1) {
	print "<BR>\n";
    }
    # Second line of buttons
    $i = 0;
    if ($config{'enable_filter'}) {
	print "<INPUT TYPE=\"submit\" NAME=\"filter\" ";
	print "VALUE=\"$config{'language'}[84]\"> ";
	print "<SELECT NAME=\"filtern\">\n";
	for ($i=0;$i<=$#{$config{'filter'}};$i++) {
	    print "<OPTION VALUE=\"$i\">";
	    print
		character_reference(${$config{'filter'}[$i]}{'description'});
	    print "\n";
	}
	print "</SELECT>\n";
	$i = 1;
    }
    if ($i == 1) {
	print "<BR>\n";
    }
    # Third line of buttons
    $i = 0;
    if ($config{'enable_upload'}) {
	print "<INPUT TYPE=\"submit\" NAME=\"upload\" ";
	print "VALUE=\"$config{'language'}[10]\">";
	$i = 1;
    }
    if ($config{'enable_mkdir'}) {
	if ($i == 1) {
	    print " ";
	}
	print "<INPUT TYPE=\"submit\" NAME=\"mkdir\" ";
	print "VALUE=\"$config{'language'}[11]\">";
	$i = 1;
    }
    if ($i == 1) {
	print "<BR>\n";
    }
    # Fourth line of buttons
    $i = 0;
    if ($config{'authentication'}) {
	print "<INPUT TYPE=\"submit\" NAME=\"logout\" ";
	print "VALUE=\"$config{'language'}[104]\">";
	$i = 1;
    }
    if ($i == 1) {
	print "<BR>\n";
    }
    # Fifth line of buttons
    $i = 0;
    if ($config{'enable_help'}) {
	print "<INPUT TYPE=\"hidden\" NAME=\"topic\" VALUE=\"panel\">";
	print "<INPUT TYPE=\"submit\" NAME=\"help\" ";
	print "VALUE=\"$config{'language'}[59]\">";
	$i = 1;
    }
    if ($i == 1) {
	print "<BR>\n";
    }
    # End
    print "</FORM>\n";
    print_foot();
}

sub get_target_directory {
    print_head();
    print "<H1>$config{'language'}[118]</H1>\n";
    print "<P><FORM ACTION=\"$config{'script_name'}\" METHOD=\"post\" ";
    print "TARGET=\"_self\">\n";
    foreach (keys %query) {
	if ($_ ne $opanel) {
	    print "<INPUT TYPE=\"hidden\" NAME=\"encoded_$_\" VALUE=\"";
	    print uri_escape_everything($query{$_});
	    print "\">\n";
	}
    }
    if (defined ($_[0])) {
	print "<P>$_[0]\n";
    }
    print "<P><INPUT TYPE=\"text\" NAME=\"$opanel\" VALUE=\"";
    print character_reference($directory{$opanel});
    print "\" SIZE=\"40\" MAXLENGTH=\"250\">\n";
    print "<P><INPUT TYPE=\"submit\" NAME=\"continue\" ";
    print "VALUE=\"$config{'language'}[119]\">\n";
    print "</FORM>\n";
    print_action_return("continue");
    print_foot();
}

sub retrieve {
    my $truepath;
    my $type;
    my $fpath;
    my $descriptions;
    my $size;
    $query{'directory'} = fix_path($query{'directory'});
    $query{'file'} = fix_path($query{'file'});
    $truepath = "$config{'base_dir'}/$query{'directory'}";
    $type = type($truepath,$query{'file'});
    $fpath = "$truepath/$query{'file'}";
    if ($config{'list_description'}) {
	$descriptions = read_descriptions($directory{$panel});
    }
    $size = (stat($fpath))[7];
    if (($type ne $config{'type_disallow'}) &&
	(sysopen(IN,$fpath,O_RDONLY)) &&
	(flock(IN,LOCK_SH))) {
	print "Content-Type: ";
	if (($type eq $config{'type_unknown'}) ||
	    ($config{'script_retrieval_always_unknown'} == 1)) {
	    print $config{'script_retrieval_unknown_mime_type'};
	}
	else {
	    print $type;
	}
	print "; name=\"$query{'file'}\"\n";
	if (($config{'list_description'}) &&
	    (defined($$descriptions{$query{'file'}}))) {
	    print "Content-Description: $$descriptions{$query{'file'}}\n";
	}
	print "Content-Disposition: inline; filename=\"$query{'file'}\"\n";
	print "Content-Length: $size\n";
	print "\n";
	print <IN>;
	flock(IN,LOCK_UN);
	close(IN);
    }
    else {
	do_error($config{'language'}[19]);
    }
}

sub get_edit {
    my $file;
    my $count = 0;
    my $realcount = 0;
    my $tmp;
    print_head();
    substitute(\$config{'language'}[74],$directory{$panel});
    print "<H1>$config{'language'}[74]</H1>\n";
    print "<P><FORM ACTION=\"$config{'script_name'}\" METHOD=\"post\" ";
    print "TARGET=\"_self\">\n";
    print_cgi_reauth();
    print_config_and_order();
    print "<INPUT TYPE=\"hidden\" NAME=\"action\" VALUE=\"edit\">\n";
    print_panel_and_directory();
    foreach $file (@{read_directory($directory{$panel})}) {
	if ($$file{'type'} ne $config{'type_disallow'}) {
	    if (defined($query{"(checked)$$file{'name'}"})) {
		$count++;
		print "<P>";
		if (-f $$file{'truepath'}) {
		    if (-T $$file{'truepath'}) {
			if ($$file{'size'} <= $config{'edit_maximum_size'}) {
			    if ((sysopen(IN,$$file{'truepath'},O_RDONLY)) &&
				(flock(IN,LOCK_SH))) {
				$realcount++;
				$tmp = $config{'language'}[78];
				substitute(\$tmp,
					   "<EM>$$file{'htmlname'}</EM>");
				print "$tmp<BR>\n";
				print "<TEXTAREA ";
				print "NAME=\"(edit)$$file{'htmlname'}\" ";
				print "ROWS=\"20\" COLS=\"80\">\n";
				while (<IN>) {
				    print character_reference($_);
				}
				flock(IN,LOCK_UN);
				close(IN);
				print "</TEXTAREA>\n";
				undef($tmp);
			    }
			    else {
				$tmp = $config{'language'}[77];
			    }
			}
			else {
			    $tmp = $config{'language'}[94];
			}
		    }
		    else {
			$tmp = $config{'language'}[76];
		    }
		}
		else {
		    $tmp = $config{'language'}[75];
		}
		if (defined($tmp)) {
		    substitute(\$tmp,"<EM>$$file{'htmlname'}</EM>");
		    print $tmp,"\n";
		    undef $tmp;
		}
	    }	
	}
    }
    if ($count == 0) {
	print "$config{'language'}[17]\n";
    }
    else {
	if ($realcount > 0) {
	    print "<P><INPUT TYPE=\"submit\" ";
	    print "VALUE=\"$config{'language'}[79]\">\n";
	}
    }
    print "</FORM>\n";
    print_action_return("edit");
    print_foot();
}

sub do_edit {
    my $file;
    my $count = 0;
    my $tmp;
    print_head();
    print "<H1>$config{'language'}[80]</H1>\n";
    print "<P>";
    foreach $file (@{read_directory($directory{$panel})}) {
	if ($$file{'type'} ne $config{'type_disallow'}) {
	    if (defined($query{"(edit)$$file{'name'}"})) {
		$count++;
		if ((-e $$file{'truepath'}) &&
		    (sysopen(OUT,$$file{'truepath'},O_WRONLY|O_TRUNC)) &&
		    (flock(OUT,LOCK_EX))) {
		    print OUT $query{"(edit)$$file{'name'}"};
		    flock(OUT,LOCK_UN);
		    close(OUT);
		    if ((length($query{"(edit)$$file{'name'}"})) !=
			(stat($$file{'truepath'}))[7]) {
			$tmp = $config{'language'}[83];
		    }
		    else {
			$tmp = $config{'language'}[81];
		    }
		}
		else {
		    $tmp = $config{'language'}[82];
		}
		substitute(\$tmp,"<EM>$$file{'htmlname'}</EM>");
		print $tmp;
		print "<BR>\n";
	    }	
	}
    }
    if ($count == 0) {
	print "$config{'language'}[17]\n";
    }
    print_action_return("");
    print_foot();
}

sub do_copy {
    sub copy {
	my $source = $_[0];
	my $target = $_[1];
	if (-e $target) {
	    return(1);
	}
	if ((sysopen(IN,$source,O_RDONLY)) &&
	    (flock(IN,LOCK_SH))) {
	    if ((sysopen(OUT,$target,O_WRONLY|O_CREAT|O_EXCL)) &&
		(flock(OUT,LOCK_EX))) {
		print OUT <IN>;
		flock(OUT,LOCK_UN);
		close(OUT);
		flock(IN,LOCK_UN);
		close(IN);
		chmod($config{'filemode'},$target);
		if ((stat($source))[7] != (stat($target))[7]) {
		    return(4);
		}
		else {
		    return(0);
		}
	    }
	    else {
		return(2);
	    }
	}
	else {
	    return(3);
	}
    }
    sub copy_recursively {
	my $file;
	my $tmp;
	my $source = $_[0];
	my $target_base = $_[1];
	my $target_rel = $_[2];
	my $target_full = fix_path("$target_base/$target_rel");
	my $truetarget = "$config{'base_dir'}/$target_full";
	my $val;
	foreach $file (@{read_directory($source)}) {
	    if ($$file{'type'} ne $config{'type_disallow'}) {
		if ($$file{'type'} eq $config{'type_directory'}) {
		    $val = mkdir("$truetarget/$$file{'name'}",0);
		    if ($val) {
			chmod($config{'dirmode'},
			      "$truetarget/$$file{'name'}");
			$tmp = $config{'language'}[20];
		    }
		    else {
			$tmp = $config{'language'}[21];
			if (-e "$truetarget/$$file{'name'}") {
			    $tmp .= " $config{'language'}[24]";
			    $val = 1;
			}
		    }
		    substitute(\$tmp,
			       "<EM>" .
			       character_reference($target_base) .
			       character_reference($target_rel) .
			       $$file{'htmlname'} .
			       "</EM>");
		}
		else {
		    $val = copy($$file{'truepath'},
				"$truetarget/$$file{'name'}");
		    if ($val == 0) {
			$tmp = $config{'language'}[64];
		    }
		    else {
			$tmp = $config{'language'}[65];	
			if ($val == 1) {
			    $tmp .= " $config{'language'}[23]";
			}
			if ($val == 2) {
			    $tmp .= " $config{'language'}[68]";
			}
			if ($val == 3) {
			    $tmp .= " $config{'language'}[67]";
			}
			if ($val == 4) {
			    $tmp .= " $config{'language'}[62]";
			}
		    }
		    substitute(\$tmp,
			       "<EM>$$file{'htmlpath'}</EM>",
			       "<EM>" .
			       character_reference($target_full) .
			       "$$file{'htmlname'}</EM>");
		}
		print $tmp;
		print "<BR>\n";
		if (($$file{'type'} eq $config{'type_directory'}) && $val) {
		    copy_recursively($$file{'path'},
				     $target_base,
				     "$target_rel$$file{'name'}/");
		}
	    }
	}
    }
    my $file;
    my $count = 0;
    my $tmp;
    my $source = $directory{$panel};
    my $target_base = $directory{$opanel};
    my $truetarget = "$config{'base_dir'}/$target_base";
    my $val;
    print_head();
    print "<H1>$config{'language'}[66]</H1>\n";
    print "<P>";
    foreach $file (@{read_directory($source)}) {
	if ($$file{'type'} ne $config{'type_disallow'}) {
	    if (defined($query{"(checked)$$file{'name'}"})) {
		$count++;
		if ($$file{'type'} eq $config{'type_directory'}) {
		    if ("$target_base$$file{'name'}/" =~
			/^$source$$file{'name'}\//) {
			$tmp = $config{'language'}[21];
			$tmp .= " $config{'language'}[63]";
			$val = 0;
		    }
		    else {
			$val = mkdir("$truetarget/$$file{'name'}",0); 
			if ($val) {
			    chmod($config{'dirmode'},
				  "$truetarget/$$file{'name'}");
			    $tmp = $config{'language'}[20];
			}
			else {
			    $tmp = $config{'language'}[21];
			    if (-e "$truetarget/$$file{'name'}") {
				$tmp .= " $config{'language'}[24]";
				$val = 1;
			    }
			}
		    }
		    substitute(\$tmp,
			       "<EM>" .
			       character_reference($target_base) .
			       $$file{'htmlname'} .
			       "</EM>");
		}
		else {
		    $val = copy($$file{'truepath'},
				"$truetarget/$$file{'name'}");
		    if ($val == 0) {
			$tmp = $config{'language'}[64];
		    }
		    else {
			$tmp = $config{'language'}[65];
			if ($val == 1) {
			    $tmp .= " $config{'language'}[23]";
			}
			if ($val == 2) {
			    $tmp .= " $config{'language'}[68]";
			}
			if ($val == 3) {
			    $tmp .= " $config{'language'}[67]";
			}
			if ($val == 4) {
			    $tmp .= " $config{'language'}[62]";
			}
		    }
		    substitute(\$tmp,
			       "<EM>$$file{'htmlpath'}</EM>",
			       "<EM>" .
			       character_reference($target_base) .
			       "$$file{'htmlname'}</EM>");
		}
		print $tmp;
		print "<BR>\n";
		if (($$file{'type'} eq $config{'type_directory'}) && $val) {
		    if ($config{'enable_copy_recursively'}) {
			copy_recursively($$file{'path'},
					 $target_base,
					 "$$file{'name'}/"),
		    }
		}
	    }	
	}
    }
    if ($count == 0) {
	print "$config{'language'}[17]\n";
    }
    print_action_return("");
    print_foot();
}

sub get_rename {
    my $file;
    my $count = 0;
    my $tmp;
    my $tmp2;
    print_head();
    substitute(\$config{'language'}[48],$directory{$panel});
    print "<H1>$config{'language'}[48]</H1>\n";
    print "<P><FORM ACTION=\"$config{'script_name'}\" METHOD=\"post\" ";
    print "TARGET=\"_self\">\n";
    print_cgi_reauth();
    print_config_and_order();
    print "<INPUT TYPE=\"hidden\" NAME=\"action\" VALUE=\"rename\">\n";
    print_panel_and_directory();
    foreach $file (@{read_directory($directory{$panel})}) {
	if ($$file{'type'} ne $config{'type_disallow'}) {
	    if (defined($query{"(checked)$$file{'name'}"})) {
		$count++;
		if ($$file{'type'} eq $config{'type_directory'}) {
		    $tmp = $config{'language'}[72];
		}
		else {
		    $tmp = $config{'language'}[73];
		}
		$tmp2 = "<INPUT TYPE=\"text\" ";
		$tmp2 .= "NAME=\"(rename)$$file{'htmlname'}\" ";
		$tmp2 .= "VALUE=\"$$file{'htmlname'}\" SIZE=\"40\" ";
		$tmp2 .= "MAXLENGTH=\"250\">";
		substitute(\$tmp,
			   "<EM>$$file{'htmlname'}</EM>",
			   $tmp2);
		print $tmp;
		print "<BR>\n";
	    }	
	}
    }
    if ($count == 0) {
	print "$config{'language'}[17]\n";
    }
    else {
	print "<INPUT TYPE=\"submit\" ";
	print "VALUE=\"$config{'language'}[49]\">\n";
    }
    print "</FORM>\n";
    print_action_return("rename");
    print_foot();
}

sub do_rename {
    my $file;
    my $count = 0;
    my $tmp;
    my $newfilename;
    my $target = "$config{'base_dir'}/$directory{$panel}";
    print_head();
    print "<H1>$config{'language'}[54]</H1>\n";
    print "<P>";
    foreach $file (@{read_directory($directory{$panel})}) {
	if ($$file{'type'} ne $config{'type_disallow'}) {
	    if (defined($query{"(rename)$$file{'name'}"})) {
		$count++;
		$newfilename = $query{"(rename)$$file{'name'}"};
		if ($newfilename eq '') {
		    $tmp = $config{'language'}[55];
		}
		elsif (! validate_name($newfilename)) {
		    $tmp = $config{'language'}[98];
		}
		elsif (rename($$file{'truepath'},"$target/$newfilename") ||
		       (system("mv",$$file{'truepath'},
			       "$target/$newfilename") == 0)) {
		    if ($$file{'type'} eq $config{'type_directory'}) {
			$tmp = $config{'language'}[50];
		    }
		    else {
			$tmp = $config{'language'}[52];
		    }
		}
		else {
		    if ($$file{'type'} eq $config{'type_directory'}) {
			$tmp = $config{'language'}[51];
		    }
		    else {
			$tmp = $config{'language'}[53];
		    }
		}
		substitute(\$tmp,
			   "<EM>$$file{'htmlpath'}</EM>",
			   "<EM>" .
			   character_reference($directory{$panel}) .
			   character_reference($newfilename) .
			   "</EM>");
		print $tmp;
		print "<BR>\n";
	    }	
	}
    }
    if ($count == 0) {
	print "$config{'language'}[17]\n";
    }
    print_action_return("");
    print_foot();
}

sub do_move {
    my $file;
    my $count = 0;
    my $tmp;
    my $target = "$config{'base_dir'}/$directory{$opanel}";
    print_head();
    print "<H1>$config{'language'}[47]</H1>\n";
    print "<P>";
    foreach $file (@{read_directory($directory{$panel})}) {
	if ($$file{'type'} ne $config{'type_disallow'}) {
	    if (defined($query{"(checked)$$file{'name'}"})) {
		$count++;
		if (rename($$file{'truepath'},"$target/$$file{'name'}") ||
		    (system("mv",$$file{'truepath'},
			    "$target/$$file{'name'}") == 0)) {
		    if ($$file{'type'} eq $config{'type_directory'}) {
			$tmp = $config{'language'}[43];
		    }
		    else {
			$tmp = $config{'language'}[45];
		    }
		}
		else {
		    if ($$file{'type'} eq $config{'type_directory'}) {
			$tmp = $config{'language'}[44];
		    }
		    else {
			$tmp = $config{'language'}[46];
		    }
		}
		substitute(\$tmp,
			   "<EM>$$file{'htmlpath'}</EM>",
			   "<EM>" .
			   character_reference($directory{$opanel}) .
			   "$$file{'htmlname'}</EM>");
		print $tmp;
		print "<BR>\n";
	    }	
	}
    }
    if ($count == 0) {
	print "$config{'language'}[17]\n";
    }
    print_action_return("");
    print_foot();
}

sub do_delete {
    sub delete_recursively {
	my $file;
	my $tmp;
	foreach $file (@{read_directory($_[0])}) {
	    if ($$file{'type'} ne $config{'type_disallow'}) {
		if ($$file{'type'} eq $config{'type_directory'}) {
		    delete_recursively($$file{'path'});
		    if (rmdir($$file{'truepath'})) {
			$tmp = $config{'language'}[13];
		    }
		    else {
			$tmp = $config{'language'}[14];
		    }
		}
		else {
		    if (unlink($$file{'truepath'})) {
			$tmp = $config{'language'}[15];
		    }
		    else {
			$tmp = $config{'language'}[16];
		    }
		}
		substitute(\$tmp,"<EM>$$file{'htmlpath'}</EM>");
		print $tmp;
		print "<BR>\n";
	    }
	}
    }
    my $file;
    my $count = 0;
    my $tmp;
    print_head();
    print "<H1>$config{'language'}[12]</H1>\n";
    print "<P>";
    foreach $file (@{read_directory($directory{$panel})}) {
	if ($$file{'type'} ne $config{'type_disallow'}) {
	    if (defined($query{"(checked)$$file{'name'}"})) {
		$count++;
		if ($$file{'type'} eq $config{'type_directory'}) {
		    if ($config{'enable_delete_recursively'}) {
			delete_recursively($$file{'path'});
		    }
		    if (rmdir($$file{'truepath'})) {
			$tmp = $config{'language'}[13];
		    }
		    else {
			$tmp = $config{'language'}[14];
		    }
		}
		else {
		    if (unlink($$file{'truepath'})) {
			$tmp = $config{'language'}[15];
		    }
		    else {
			$tmp = $config{'language'}[16];
		    }
		}
		substitute(\$tmp,"<EM>$$file{'htmlpath'}</EM>");
		print $tmp;
		print "<BR>\n";
	    }	
	}
    }
    if ($count == 0) {
	print "$config{'language'}[17]\n";
    }
    print_action_return("");
    print_foot();
}

sub get_describe {
    my $descriptions;
    my $file;
    my $count = 0;
    my $tmp;
    my $description;
    my $tmp2;
    print_head();
    substitute(\$config{'language'}[58],$directory{$panel});
    print "<H1>$config{'language'}[58]</H1>\n";
    print "<P><FORM ACTION=\"$config{'script_name'}\" METHOD=\"post\" ";
    print "TARGET=\"_self\">\n";
    print_cgi_reauth();
    print_config_and_order();
    print "<INPUT TYPE=\"hidden\" NAME=\"action\" VALUE=\"describe\">\n";
    print_panel_and_directory();
    $descriptions = read_descriptions($directory{$panel});
    foreach $file (@{read_directory($directory{$panel})}) {
	if ($$file{'type'} ne $config{'type_disallow'}) {
	    if (defined($query{"(checked)$$file{'name'}"})) {
		$count++;
		if ($$file{'type'} eq $config{'type_directory'}) {
		    $tmp = $config{'language'}[111];
		}
		else {
		    $tmp = $config{'language'}[112];
		}
		if (defined($$descriptions{$$file{'name'}})) {
		    $description =
			character_reference($$descriptions{$$file{'name'}});
		}
		else {
		    $description = '';
		}
		$tmp2 = "<INPUT TYPE=\"text\" ";
		$tmp2 .= "NAME=\"(describe)$$file{'htmlname'}\" ";
		$tmp2 .= "VALUE=\"$description\" SIZE=\"40\" ";
		$tmp2 .= "MAXLENGTH=\"511\">";
		substitute(\$tmp,
			   "<EM>$$file{'htmlname'}</EM>",
			   $tmp2);
		print $tmp;
		print "<BR>\n";
	    }	
	}
    }
    if ($count == 0) {
	print "$config{'language'}[17]\n";
    }
    else {
	print "<INPUT TYPE=\"submit\" ";
	print "VALUE=\"$config{'language'}[113]\">\n";
    }
    print "</FORM>\n";
    print_action_return("describe");
    print_foot();
}

sub do_describe {
    my $descriptions;
    my $file;
    my $count = 0;
    my $tmp;
    my $newfilename;
    my $target = "$config{'base_dir'}/$directory{$panel}";
    print_head();
    print "<H1>$config{'language'}[114]</H1>\n";
    print "<P>";
    $descriptions = read_descriptions($directory{$panel});
    foreach $file (@{read_directory($directory{$panel})}) {
	if ($$file{'type'} ne $config{'type_disallow'}) {
	    if (defined($query{"(describe)$$file{'name'}"})) {
		$count++;
		$$descriptions{$$file{'name'}} =
		    $query{"(describe)$$file{'name'}"};
	    }
	}
    }
    if ($count == 0) {
	print $config{'language'}[17];
    }
    elsif (write_descriptions($directory{$panel},
			      $descriptions)) {
	print $config{'language'}[115];
    }
    else {
	print $config{'language'}[116];
    }
    print_action_return("");
    print_foot();
}

sub do_filter {
    my $iterative = 0;
    my $sequence = 0;
    my $tmp;
    my $command = ${$config{'filter'}[$query{'filtern'}]}{'command'};
    my $description = ${$config{'filter'}[$query{'filtern'}]}{'description'};
    my $file;
    my $count = 0;
    my @command;
    my @list = ();
    my @tmp;
    print_head();
    print "<H1>$config{'language'}[85]</H1>\n";
    print "<P>";
    if ($command =~ /!s!/) {
	$iterative = 1;
    }
    if ($command =~ /!l!/) {
	$sequence = 1;
    }
    $tmp = $config{'base_dir'} . $directory{$panel};
    chdir($tmp);
    if ($command =~ /!t!/) {
	$command =~ s/!t!/$tmp/go;
    }
    if ($command =~ /!o!/) {
	$tmp = "drall-";
	$tmp .= time();
	print "<STRONG>";
	substitute(\$config{'language'}[92],
		   "<EM>" . character_reference($description) . "</EM>",
		   "<EM>$tmp</EM>");
	print $config{'language'}[92];
	print "</STRONG>\n";
	print "<P>";
	$command =~ s/!o!/$tmp/go;
    }
    foreach $file (@{read_directory($directory{$panel})}) {
	if ($$file{'type'} ne $config{'type_disallow'}) {
	    if (defined($query{"(checked)$$file{'name'}"})) {
		$count++;
		if ($sequence == 1) {
		    $list[$#list+1] = $$file{'name'};
		}
		if ($iterative == 1) {
		    @command = split(/ /, $command);
		    local $_;
		    foreach (@command) {
			s/!s!/$$file{'name'}/go;
		    }
		    if (system(@command) == 0) {
			if ($$file{'type'} eq $config{'type_directory'}) {
			    $tmp = $config{'language'}[86];
			}
			else {
			    $tmp = $config{'language'}[87];
			}
		    }
		    else {
			if ($$file{'type'} eq $config{'type_directory'}) {
			    $tmp = $config{'language'}[88];
			}
			else {
			    $tmp = $config{'language'}[89];
			}
		    }
		    substitute(\$tmp,
			       "<EM>" .
			       character_reference($description) .
			       "</EM>",
			       "<EM>$$file{'htmlpath'}</EM>");
		    print $tmp;
		    print "<BR>\n";
		}
	    }	
	}
    }
    if ($count == 0) {
	print "$config{'language'}[17]\n";
    }
    else {
	if ($sequence == 1) {
	    @command = ();
	    @tmp = split(/ /, $command);
	    local $_;
	    foreach (@tmp) {
		if ($_ eq '!l!') {
		    local $_;
		    foreach (@list) {
			$command[$#command+1] = $_;
		    }
		}
		else {
		    $command[$#command+1] = $_;
		}
	    }
	    if (system(@command) == 0) {
		$tmp = $config{'language'}[90];
	    }
	    else {
		$tmp = $config{'language'}[91];
	    }
	    substitute(\$tmp,
		       "<EM>" .
		       character_reference($description) .
		       "</EM>");
	    print $tmp;
	    print "<BR>\n";
	}
    }
    print_action_return("");
    print_foot();
}

sub get_upload {
    print_head();
    if ($_[0] > 0) {
	print "<H1>$config{'language'}[37]</H1>\n";
	print "<P>";
	if (($_[0] == 1) || ($_[0] == 6) || ($_[0] == 8) ||
	    ($_[0] == 9) || ($_[0] == 10) || ($_[0] == 11)) {
	    substitute(\$config{'language'}[27],
		       "<EM>" .
		       character_reference($query{'filename'}) .
		       "</EM>",
		       "<EM>" .
		       character_reference($directory{$panel}) .
		       "</EM>");
	    print $config{'language'}[27];
	}
	elsif ($_[0] != 5) {
	    substitute(\$config{'language'}[28],
		       "<EM>" .
		       character_reference($query{'filename'}) .
		       "</EM>",
		       "<EM>" .
		       character_reference($directory{$panel}) .
		       "</EM>");
	    print $config{'language'}[28];
	}
	if ($_[0] == 2) {
	    print " $config{'language'}[29]";
	}
	elsif ($_[0] == 3) {
	    print " $config{'language'}[30]";
	}
	elsif ($_[0] == 4) {
	    print " $config{'language'}[31]";
	}
	elsif ($_[0] == 5) {
	    print $config{'language'}[32];
	}
	if (($_[0] == 6) || ($_[0] == 8) || ($_[0] == 10)) {
	    print " $config{'language'}[39]";
	}
	if ($_[0] == 7) {
	    print " $config{'language'}[97]";
	}
	if (($_[0] == 8) || ($_[0] == 9)) {
	    print " $config{'language'}[109]";
	}
	if (($_[0] == 10) || ($_[0] == 11)) {
	    print " $config{'language'}[110]";
	}
	print "\n";
	print "<P><HR>\n";
    }
    substitute(\$config{'language'}[33],$directory{$panel});
    print "<H1>$config{'language'}[33]</H1>\n";
    print "<P><FORM ENCTYPE=\"multipart/form-data\" ";
    print "ACTION=\"$config{'script_name'}\" METHOD=\"post\" ";
    print "TARGET=\"_self\">\n";
    print_cgi_reauth();
    print_config_and_order();
    print "<INPUT TYPE=\"hidden\" NAME=\"action\" VALUE=\"upload\">\n";
    print_panel_and_directory();
    print "<INPUT TYPE=\"checkbox\" NAME=\"lowercase\" CHECKED>";
    print " $config{'language'}[34]<BR>\n";
    print "<INPUT TYPE=\"checkbox\" NAME=\"overwrite\">";
    print " $config{'language'}[38]<BR>\n";
    print "<INPUT TYPE=\"file\" NAME=\"file\" SIZE=\"40\" ";
    print "MAXLENGTH=\"250\"><BR>\n";
    if ($config{'enable_upload_description'}) {
	print "<INPUT TYPE=\"text\" NAME=\"description\" SIZE=\"40\" ";
	print "MAXLENGTH=\"511\"> $config{'language'}[108]<BR>\n";
    }
    print "<INPUT TYPE=\"submit\" VALUE=\"$config{'language'}[35]\">\n";
    print "</FORM>\n";
    print_action_return("upload");
    print_foot();
}

sub do_upload {
    my $fpath;
    my $overwritten = 0;
    my $descriptions;
    $query{'filename'} = fix_path($query{'filename'});
    if (defined($query{'lowercase'})) {
	$query{'filename'} = lc($query{'filename'});
    }
    $fpath = $config{'base_dir'} . $directory{$panel} . $query{'filename'};
    if ($query{'filename'} eq '') {
	get_upload(5);
    }
    elsif (! validate_name($query{'filename'})) {
	get_upload(7);
    }
    elsif (-d $fpath) {
	get_upload(3);
    }
    elsif ((-e $fpath) && (! defined($query{'overwrite'}))) {
	get_upload(4);
    }
    else {
	if ((-e $fpath) && (defined($query{'overwrite'}))) {
	    $overwritten = 1;
	}
	if ((sysopen(OUT,$fpath,O_WRONLY|O_CREAT|O_TRUNC)) &&
	    (flock(OUT,LOCK_EX))) {
	    print OUT $query{'file'};
	    flock(OUT,LOCK_UN);
	    close(OUT);
	    chmod($config{'filemode'},$fpath);
	    if (($config{'enable_upload_description'}) &&
		($query{'description'} ne '')) {
		$descriptions = read_descriptions($directory{$panel});
		$$descriptions{$query{'filename'}} = $query{'description'};
		if (write_descriptions($directory{$panel},
				       $descriptions)) {
		    if ($overwritten == 1) {
			get_upload(8);
		    }
		    else {
			get_upload(9);
		    }    
		}
		else {
		    if ($overwritten == 1) {
			get_upload(10);
		    }
		    else {
			get_upload(11);
		    }    
		}
	    }
	    else {
		if ($overwritten == 1) {
		    get_upload(6);
		}
		else {
		    get_upload(1);
		}
	    }
	}
	else {
	    get_upload(2);
	}
    }
}

sub get_mkdir {
    print_head();
    if ($_[0] > 0) {
	print "<H1>$config{'language'}[36]</H1>\n";
	print "<P>";
	if ($_[0] == 1) {
	    substitute(\$config{'language'}[20],
		       "<EM>" .
		       character_reference($directory{$panel}) .
		       character_reference($query{'directory'}) .
		       "</EM>");
	    print $config{'language'}[20];
	}
	else {
	    substitute(\$config{'language'}[21],
		       "<EM>" .
		       character_reference($directory{$panel}) .
		       character_reference($query{'directory'}) .
		       "</EM>");
	    print $config{'language'}[21];
	}
	if ($_[0] == 2) {
	    print " $config{'language'}[22]";
	}
	elsif ($_[0] == 3) {
	    print " $config{'language'}[23]";
	}
	elsif ($_[0] == 4) {
	    print " $config{'language'}[24]";
	}
	elsif ($_[0] == 5) {
	    print " $config{'language'}[56]";
	}
	elsif ($_[0] == 6) {
	    print " $config{'language'}[96]";
	}
	print "\n";
	print "<P><HR>\n";
    }
    substitute(\$config{'language'}[25],$directory{$panel});
    print "<H1>$config{'language'}[25]</H1>\n";
    print "<P><FORM ACTION=\"$config{'script_name'}\" METHOD=\"post\" ";
    print "TARGET=\"_self\">\n";
    print_cgi_reauth();
    print_config_and_order();
    print "<INPUT TYPE=\"hidden\" NAME=\"action\" VALUE=\"mkdir\">\n";
    print_panel_and_directory();
    print "<INPUT TYPE=\"text\" NAME=\"directory\" SIZE=\"40\" ";
    print "MAXLENGTH=\"250\"><BR>\n";
    print "<INPUT TYPE=\"submit\" VALUE=\"$config{'language'}[26]\">\n";
    print "</FORM>\n";
    if (($config{'return_to_maked_directory'}) && ($_[0] == 1)) {
	$directory{$panel} .= $query{'directory'};
    }
    print_action_return("mkdir");
    print_foot();
}

sub do_mkdir {
    my $dpath;
    $query{'directory'} = fix_path($query{'directory'});
    $dpath = $config{'base_dir'} . $directory{$panel} . $query{'directory'};
    if ($query{'directory'} eq '') {
	get_mkdir(5);
    }
    elsif (! validate_name($query{'directory'})) {
	get_mkdir(6);
    }
    elsif (-f $dpath) {
	get_mkdir(3);
    }
    elsif (-e $dpath) {
	get_mkdir(4);
    }
    else {
	if (mkdir($dpath,0)) {
	    chmod($config{'dirmode'},$dpath);
	    get_mkdir(1);
	}
	else {
	    get_mkdir(2);
	}
    }
}

sub do_error {
    if ($config{'authentication'}) {
	Averist::logout();
	undef $username;
	undef $ticket;
    }
    print_error(@_);
}

sub print_error {
    my $line;
    print_head();
    print "<H1 ALIGN=\"center\">$config{'language'}[117]</H1>\n";
    print "<DIV ALIGN=\"center\">\n";
    foreach $line (@_) {
	print $line;
	print "<BR>\n";
    }
    print "</DIV>\n";
    print_foot();
}

###########################
# Interface Functions End #
###########################

#####################  Functions used internally to service other functions.
# Service Functions #
#####################

sub fix_config {
    my $file;
    if ($config{'config_multi'} >= 1) {
	if (defined($query{'config'})) {
	    $query{'config'} = fix_path($query{'config'});
	    $file = $config{'config_directory'} .
		"$query{'config'}.external.drall.pl";
	    if (-r $file) {
		require $file;
	    }
	    else {
		print_error($config{'language'}[40]);
		return(0);
	    }
	}
	elsif ($config{'config_multi'} >=2) {
	    print_error($config{'language'}[41]);
	    return(0);
	}
    }
    return(1);
}

sub check_ssl {
    if (($config{'ssl'} == 1) && 
	((! defined($ENV{'HTTPS'})) && # Apache (and maybe others)
	 (! defined($ENV{'SUPPORTS_ssl'})))) { # Roxen Challenger/WebServer
	print_error($config{'language'}[106]);
	return(0);
    }
    return(1);
}

sub authenticate {
    if ($config{'authentication'}) {
	require Averist;
	if ($Averist::Config{'events'}{'authentication'}{'remote'} ne 'cgi') {
	    print_error($config{'language'}[57]);
	    return(0);
	}
	if ((defined($query{'action'})) &&
	    (($query{'action'} eq 'list') ||
	     ($query{'action'} eq 'retrieve'))) {
	    ($username,$ticket) = Averist::login(0,\%query);
	}
	else {
	    ($username,$ticket) = Averist::login(1,\%query);
	}
	if (! defined($ticket)) {
	    print_login();
	    return(0);
	}
    }
    else {
	if (defined($ENV{'REMOTE_USER'})) {
	    $username = $ENV{'REMOTE_USER'};
	}
    }
    return(1);
}

sub suid {
    my $uid;
    my $gid;
    if ((defined($username)) && ($config{'suid'})) {
	(undef,undef,$uid,$gid) = getpwnam($username);
	if ((! defined($gid)) || (! defined($uid))) {
	    do_error($config{'language'}[107]);
	    return(0);
	}
	$) = "$gid $gid"; 
	$> = $uid; 
    }
    return(1);
}

sub substitute_recursive {
    my $struct = $_[0];
    my $var = $_[1];
    my $val = $_[2];
    my $i;
    if (scalar($struct)) {
	$struct =~ s/!$var!/$val/go;
    }
    if (ref($struct) eq "ARRAY") {
	for ($i=0;$i<@{$struct};$i++) {
	    ${$struct}[$i] = substitute_recursive(${$struct}[$i],$var,$val);
	}
    }
    if (ref($struct) eq "HASH") {
	foreach (keys %{$struct}) {
	    ${$struct}{$_} = substitute_recursive(${$struct}{$_},$var,$val);
	}
    }
    return($struct);
}

sub validate_name {
    my $filename = $_[0];
    if ($filename =~ /[\\\/]/) {
	return(0);
    }
    else {
	return(1);
    }
}

sub expand_symlink {
    my ($dir,$link) = @_;
    my $linkpart;
    $dir =~ s-/+-/-g;
    $link =~ s-/+-/-g;
    $dir =~ s-/$--g;
    $link =~ s-/$--g;
    my @dirs = split(/\//,$dir);
    my @links = split(/\//,$link);
    while (defined($linkpart = shift(@links))) {
	if ($linkpart eq '') {
	    @dirs = ('');
	}
	elsif ($linkpart eq '.') {
	    # Do nothing
	} 
	elsif ($linkpart eq '..') {
	    if ($#dirs) { pop(@dirs); }
	} 
	else {
	    push(@dirs,$linkpart);
	}
    }
    return(join('/',@dirs,''));
}

sub validate_path {
    my $counter = 0;
    my $resolved;
    my @dirs;
    my $dir = $_[0];
    my $file = $_[1];
    while (-l "$dir$file") {
	if ($counter++ == 100) {
	    return(0); # Too many symbolic links, make dir/file invalid
	}
	$file = readlink("$dir$file");
	$resolved = expand_symlink($dir,$file);
	chop($resolved);
	@dirs = split(/\//,$resolved);
	$file = $dirs[$#dirs];
	$dirs[$#dirs] = '';
	$dir = join('/',@dirs,'');
	chop($dir);
    }
    if (-d "$dir$file") {
	$file .= "/";
    }
    elsif ($config{'symlink_security'} == 0) {
	return(1);
    }
    elsif ("$dir$file" =~ /^$config{'base_dir'}/) {
	return(1);
    }
    else {
	return(0);
    }
}

sub validate_directory {
    if (! -d "$_[0]$_[1]") {
	return(0);
    }
    elsif (validate_path($_[0],$_[1])) {
	return(1);
    }
    else {
	return(0);
    }
}

sub fix_path {
    my $path = $_[0];
    $path =~ s{^\.\./}{/};
    $path =~ s{/\.\.$}{/};
    while ($path =~ /\/\.\.\//) {
	$path =~ s{/\.\./}{/}g;
    }
    while ($path =~ /\/\.\//) {
	$path =~ s{/\./}{/}g;
    }
    $path =~ s{/+}{/}g;
    $path =~ s{http:/}{http://}i;
    $path =~ s{https:/}{https://}i;
    return($path);
}

sub uri_escape {
    my $tmp = $_[0];
    $tmp =~ s{([^\w\$\@./:-])}{sprintf("%%%02x",ord($1))}gxe;
    return($tmp);
}

sub underscore {
    my $tmp = $_[0];
    $tmp =~ s{[^\w\$\@./:-]}{_}g;
    return($tmp);
}

sub uri_escape_everything {
    my $tmp = $_[0];
    $tmp =~ s{(.)}{sprintf("%%%02x",ord($1))}gxes;
    return($tmp);
}

sub character_reference {
    my $string = $_[0];
    $string =~ s/\&/\&amp\;/g;
    $string =~ s/\</\&lt\;/g;
    $string =~ s/\>/\&gt\;/g;
    $string =~ s/\"/\&quot\;/g;
    return($string);
}

sub substitute {
    my $string = shift;
    if ((ref($string)) && (scalar($$string))) {
	$$string =~ s/!(\d+?)!/$_[$1]/eg;
    }
    elsif (scalar($string)) {
	$string =~ s/!(\d+?)!/$_[$1]/eg;
    }
    return $string;
}

sub read_directory {
    my $path = $_[0];
    my $truepath = fix_path("$config{'base_dir'}$path");
    my $url = fix_path("$config{'base_url'}$path");
    my $i;
    my $direntry;
    my @stat;
    my @dirstruct;
    my @empty = ();
    my @current_directory_listing;
    my $sorted = 0;
    if (opendir(DIR,$truepath)) {
	for ($i=0;defined($direntry=readdir(DIR));$i++) {
	    @stat = stat("$truepath/$direntry");
	    $dirstruct[$i] = {
		'name'     => $direntry,
		'htmlname' => character_reference($direntry),
		'path'     => "$path$direntry",
		'htmlpath' => character_reference("$path$direntry"),
		'truepath' => "$truepath$direntry",
		'url'      => uri_escape("$url$direntry"),
		'type'     => type($truepath,$direntry),
		'size'     => $stat[7],
		'time'     => $stat[9],
	    };
	    if ($dirstruct[$i]{'type'} eq $config{'type_directory'}) {
		$dirstruct[$i]{'path'} .= '/';
		$dirstruct[$i]{'truepath'} .= '/';
		$dirstruct[$i]{'url'} .= '/';
	    }
	}
	closedir(DIR);
    }
    else {
	return(\@empty);
    }
    if (defined($query{"order-$panel"})) {
	$sorted = 1;
	if ($query{"order-$panel"} eq 'name-reverse') {
	    @current_directory_listing = sort {
		($$b{'type'} eq $config{'type_directory'}) <=>
		    ($$a{'type'} eq $config{'type_directory'}) ||
			$$b{'name'} cmp $$a{'name'} } @dirstruct;
	}
	elsif ($query{"order-$panel"} eq 'size') {
	    @current_directory_listing = sort {
		($$b{'type'} eq $config{'type_directory'}) <=>
		    ($$a{'type'} eq $config{'type_directory'}) ||
			$$a{'size'} <=> $$b{'size'} } @dirstruct;
	}
	elsif ($query{"order-$panel"} eq 'size-reverse') {
	    @current_directory_listing = sort {
		($$b{'type'} eq $config{'type_directory'}) <=>
		    ($$a{'type'} eq $config{'type_directory'}) ||
			$$b{'size'} <=> $$a{'size'} } @dirstruct;
	}
	elsif ($query{"order-$panel"} eq 'time') {
	    @current_directory_listing = sort {
		($$b{'type'} eq $config{'type_directory'}) <=>
		    ($$a{'type'} eq $config{'type_directory'}) ||
			$$a{'time'} <=> $$b{'time'} } @dirstruct;
	}
	elsif ($query{"order-$panel"} eq 'time-reverse') {
	    @current_directory_listing = sort {
		($$b{'type'} eq $config{'type_directory'}) <=>
		    ($$a{'type'} eq $config{'type_directory'}) ||
			$$b{'time'} <=> $$a{'time'} } @dirstruct;
	}
	else {
	    $sorted = 0;
	}
    }
    if ($sorted == 0) {
	@current_directory_listing = sort {
	    ($$b{'type'} eq $config{'type_directory'}) <=>
		($$a{'type'} eq $config{'type_directory'}) ||
		    $$a{'name'} cmp $$b{'name'} } @dirstruct;
    }
    return(\@current_directory_listing);
}

sub read_descriptions {
    my %descriptions;
    if ((sysopen(DESCRIPTIONS,
		 "$config{'base_dir'}/$_[0]/$config{'description_filename'}",
		 O_RDONLY)) &&
	(flock(DESCRIPTIONS,LOCK_SH))) {
	while (<DESCRIPTIONS>) {
	    if ((/^"(.+)"\s+(.+)$/) || (/^(\S+)\s+(.+)$/)) {
		$descriptions{$1} = $2;
	    }
	}
	flock(DESCRIPTIONS,LOCK_UN);
	close(DESCRIPTIONS);
    }
    return(\%descriptions);
}

sub write_descriptions {
    my $descriptions = $_[1];
    my $fpath = "$config{'base_dir'}/$_[0]/$config{'description_filename'}";
    if ((sysopen(DESCRIPTIONS,$fpath,O_WRONLY|O_CREAT|O_TRUNC)) &&
	(flock(DESCRIPTIONS,LOCK_EX))) {
	foreach (sort(keys(%{$descriptions}))) {
	    @_ = split(/\./);
	    if ((! /\s/) ||
		((@_ == 1) &&
		 (length($_[0]) <= 8)) ||
		((@_ == 2) &&
		 (length($_[0]) <= 8) &&
		 (length($_[1]) <= 3))) {
		print DESCRIPTIONS "$_ $$descriptions{$_}\n";
	    }
	    else {
		print DESCRIPTIONS "\"$_\" $$descriptions{$_}\n";
	    }
	}
	flock(DESCRIPTIONS,LOCK_UN);
	close(DESCRIPTIONS);
	chmod($config{'filemode'},$fpath);
	return(1);
    }
    return(0);
}

sub type {
    my $directory = $_[0];
    my $file = $_[1];
    my $type;
    if (($file eq '.') ||
	($file eq '..') ||
	(! validate_name($file)) ||
	(! validate_path($directory,$file))) {
	$type = $config{'type_disallow'};
    }
    else {
	foreach (@{$config{'hide_files_and_directories'}}) {
	    if ($file =~ /$_/) {
		$type = $config{'type_hidden'};
	    }
	}
	if (! defined($type)) {
	    if (-d "$directory/$file") {
		$type = $config{'type_directory'};
	    }
	    elsif ($file =~ /\.(\w+)$/) {
		if (defined($config{'ext2type'}{$1})) {
		    $type = $config{'ext2type'}{$1};
		}
		elsif (defined($config{'ext2type'}{lc($1)})) {
		    $type = $config{'ext2type'}{lc($1)};
		}
	    }
	}
    }
    if (! defined($type)) {
	$type = $config{'type_unknown'};
    }
    return($type);
}

#########################
# Service Functions End #
#########################

#################  Function layer for the Common Gateway Interface.
# CGI Functions #  These could possible be altered to fit other gateway
#################  interfaces if such appear in the future.

sub parse_query {
    if ((defined($ENV{'CONTENT_TYPE'})) &&
	($ENV{'CONTENT_TYPE'} =~ /multipart\/form-data/i)) {
	return(parse_multipart_query());
    }
    else {
	return(parse_normal_query());
    }
}

sub parse_multipart_query {
    ## If a file is sent (input type "file") its name will be associated with
    ## "filename" in the returned associative array.
    my @input;
    my $multipart_divider;
    my $temp_line;
    my $temp_head;
    my $temp_fieldname;
    my $temp_filename;
    my $temp_body;
    my $temp_data;
    my %in;
    my @temp_tmp;
    @input = <STDIN>;
    $multipart_divider = $input[0];
    $multipart_divider =~ s/\r*\n*$//;
    $temp_line = 1;
    while (1) {
	$temp_head = $input[$temp_line++];
	last if ($temp_line > $#input);
	($temp_fieldname) = ($temp_head =~ / name=\"([^\"]+)\"/);
	if (! defined($temp_filename)) {
	    ($temp_filename)  = ($temp_head =~ / filename=\"([^\"]+)\"/);
	}
	while ($temp_head = $input[$temp_line++]) {
	    last if ($temp_head =~ /^\s*\r*\n*$/);
	}
	$temp_body = '';
	while ($temp_data = $input[$temp_line++]) {
	    last if ($temp_data =~ /^$multipart_divider/);
	    last if ($temp_line > $#input);
	    $temp_body .= $temp_data;
	}
	$temp_body =~ s/\r*\n*$//;
	$in{$temp_fieldname} = $temp_body;
	last if ($temp_line > $#input);
    }
    $in{'filename'} = $temp_filename;
    if ($in{'filename'} =~ /\\/) {
	@temp_tmp = split(/\\/,$in{'filename'});	
	$in{'filename'} = $temp_tmp[$#temp_tmp];
    }
    if ($in{'filename'} =~ /\//) {
	@temp_tmp = split(/\//,$in{'filename'});	
	$in{'filename'} = $temp_tmp[$#temp_tmp];
    }
    return(%in);
}

sub parse_normal_query {
    my $in;
    my @in;
    my $key;
    my $val;
    my %in;
    if (defined($ENV{'REQUEST_METHOD'})) {
	if ($ENV{'REQUEST_METHOD'} eq 'GET') {
	    $in = $ENV{'QUERY_STRING'};
	}
	elsif ($ENV{'REQUEST_METHOD'} eq 'POST') {
	    if (defined($ENV{'CONTENT_LENGTH'})) {
		read(STDIN,$in,$ENV{'CONTENT_LENGTH'});
	    }
	    else {
		return();
	    }
	}
	else {
	    return();
	}
    }
    else {
	return();
    }
    @in = split(/[&;]+/,$in);
    foreach (@in) {
	s/\+/ /g;
	($key,$val) = split(/=/);
	$key =~ s/%(..)/pack("c",hex($1))/ge;
	$val =~ s/%(..)/pack("c",hex($1))/ge;
	$key =~ s/\0//g;
	$val =~ s/\0//g;
	if ($key =~ /^encoded_(.*)$/) {
	    $key = $1;
	    $val =~ s/%(..)/pack("c",hex($1))/ge;
	}
	$in{$key} = $val;
    }
    return(%in);
}

#####################
# CGI Functions End #
#####################


#################
# Functions End #
#################



##############
# Script End #
##############
